import os
import logging
import uuid
from typing import Callable
from concurrent.futures import ThreadPoolExecutor  # ,ProcessPoolExecutor
from kgcnn.mol.external.ballloon import BalloonInterface
from kgcnn.mol.io import read_mol_list_from_sdf_file, write_smiles_file

logging.basicConfig()  # Module logger
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)

try:
    from kgcnn.mol.module_rdkit import convert_smile_to_mol_rdkit as rdkit_smile_to_mol
except ImportError:
    module_logger.error("Can not import RDKit module for conversion.")
    rdkit_smile_to_mol = None

try:
    from kgcnn.mol.module_babel import convert_smile_to_mol_openbabel as openbabel_smile_to_mol
    # There are problems with openbabel if system variable is not set.
    # Openbabel may not be fully threadsafe, but is improved in version 3.0.
    if "BABEL_DATADIR" not in os.environ:
        module_logger.warning(
            "In case openbabel fails, you can set `kgcnn.mol.convert.openbabel_smile_to_mol` to `None` for disable.")
except ImportError:
    module_logger.error("Can not import OpenBabel module for conversion.")
    openbabel_smile_to_mol = None


class MolConverter:

    def __init__(self,
                 base_path: str = None,
                 external_program: dict = None,
                 num_workers: int = None,
                 sanitize: bool = True,
                 add_hydrogen: bool = True,
                 make_conformers: bool = True,
                 optimize_conformer: bool = True):
        """Initialize a converter to transform smile or coordinates into mol block information.

        Args:
            base_path (str):
            external_program (dict):
            num_workers (int):
            sanitize (bool):
            add_hydrogen (bool):
            make_conformers (bool):
            optimize_conformer (bool):
        """
        self.base_path = base_path
        self.external_program = external_program
        self.num_workers = num_workers
        self.sanitize = sanitize
        self.add_hydrogen = add_hydrogen
        self.make_conformers = make_conformers
        self.optimize_conformer = optimize_conformer

        if base_path is None:
            self.base_path = os.path.realpath(__file__)
        if num_workers is None:
            self.num_workers = os.cpu_count()

    @staticmethod
    def _check_is_correct_length(a, b):
        if len(a) != len(b):
            module_logger.error("Mismatch in number of converted. Found %s vs. %s" % (len(a), len(b)))
            raise ValueError("Conversion was not successful")

    @staticmethod
    def _convert_parallel(conversion_method: Callable,
                          smile_list: list,
                          num_workers: int,
                          *args
                          ):
        if num_workers is None:
            num_workers = os.cpu_count()

        if num_workers == 1:
            mol_list = [conversion_method(x, *args) for x in smile_list]
            return mol_list
        else:
            arg_list = [(x,) + args for x in smile_list]
            with ThreadPoolExecutor(max_workers=num_workers) as executor:
                result = executor.map(conversion_method, *zip(*arg_list))
            mol_list = list(result)
            return mol_list

    @staticmethod
    def _single_smile_to_mol(smile: str,
                             sanitize: bool = True,
                             add_hydrogen: bool = True,
                             make_conformers: bool = True,
                             optimize_conformer: bool = True):
        if rdkit_smile_to_mol is not None:
            mol = rdkit_smile_to_mol(smile=smile, sanitize=sanitize, add_hydrogen=add_hydrogen,
                                     make_conformers=make_conformers, optimize_conformer=optimize_conformer)
            if mol is not None:
                return mol

        if openbabel_smile_to_mol is not None:
            mol = openbabel_smile_to_mol(smile=smile, sanitize=sanitize, add_hydrogen=add_hydrogen,
                                         make_conformers=make_conformers, optimize_conformer=optimize_conformer)
            if mol is not None:
                return mol

        module_logger.warning("Failed conversion for smile %s" % smile)
        return None

    def smile_to_mol(self, smile_list: list):

        if self.external_program is None:
            # Default via rdkit and openbabel
            mol_list = self._convert_parallel(self._single_smile_to_mol,
                                              smile_list,
                                              self.num_workers,
                                              self.sanitize,
                                              self.add_hydrogen,
                                              self.make_conformers,
                                              self.optimize_conformer)
            # Check success
            self._check_is_correct_length(smile_list, mol_list)
            return mol_list

        # External programs

        # Write out temporary smiles file.
        smile_file = os.path.join(self.base_path, str(uuid.uuid4()) + ".smi")
        mol_file = os.path.splitext(smile_file)[0] + ".sdf"

        write_smiles_file(smile_file, smile_list)

        if self.external_program["class_name"] == "balloon":
            ext_program = BalloonInterface(**self.external_program["config"])
            ext_program.run(input_file=smile_file, output_file=mol_file, output_format="sdf")
        else:
            raise ValueError("Unknown program for conversion of smiles %s" % self.external_program)

        mol_list = read_mol_list_from_sdf_file(mol_file)
        # Clean up
        os.remove(mol_file)
        os.remove(smile_file)

        # Check success
        self._check_is_correct_length(smile_list, mol_list)
        return mol_list



import sys
import os
import subprocess


class BalloonInterface:
    """Interface to Balloon program. This is tested for Balloon version 1.8.1.
    Described in http://dx.doi.org/10.1021/ci6005646
    Copyright (C) 2006-2021 Mikko J. Vainio and J. Santeri Puranen.
    Copyright (C) 2010 Visipoint Ltd. www.visipoint.fi.

    Setup Balloon configuration. This is the help parameter option of balloon help output mapped to python.
    The parameters are named identical except of `input_file`, `output_file` and `output_format`.

    Args:
        config (str): Name of a configuration file to read. Command-line options override the ones in the
            config file. A config file may contain lines that look like 'long_option_name = value'
            and comment lines that begin with '#'. Default is None.
        writeMOL2 (bool): Force output of structures in MOL2 format (can store partial atomic charges). See also
            'output-format'. Default is False.
        onlycharge (bool): Only assign partial atomic charges to the input structures and output in MOL2 format. A
            shortcut option for -c0 -k --writeMOL2. See also 'chargemodel'. Default is False.
        nobadmodels (bool): Do not write bad models to '<output-file>_bad<.suffix>'. Default is False.
        strict (bool): Skip structures that cannot be handled by the used force field. Default is False.
        nconfs (int): Number of conformers to generate or the initial population size if using GA. Zero
            will cause the input structure to be written out with partial atomic charges and energy as
            calculated by the MMFF94-like force field. Default is 1.
        randomSeed (int): Seed the pseudo-random number generator. Range [1, 4294967295). Default value taken
            from clock. Default is None.
        sdfnamefieldheader (str): When input is in MDL SD file format, use the contents of the data field with the
            given header as the molecule name. Only the first line of the data is considered.
        output_file (str): Name for the output file. If not given, the last filename in the input list is taken as
            the output file name. The file format is deduced from the file extension if not explicitly forced using
            'output-format'. Recognized extensions are sdf, sdf:v3, mol2, smi, vbf, xyz.
        output_format (str): Force output in the given format. Format is one of
            - sdf: Accelrys mol/sdf format.
            - sdf(v3): Accelrys mol/sdf format (V3000)
            - mol2: Tripos Mol2 format
            - smi: SMILES format
            - vbf: Visipoint Binary Format (experimental)
            - xyz: XMol XYZ format (experimental code)
        input_file (str): Input file name. File lists from multiple occurrences of the option are concatenated.
            If no output file is provided, the last name in the input file list is used as the output file name.
            Input file format is recognized automatically from the file contents. Recognized formats are
            - sdf: Accelrys mol/sdf format.
            - sdf(v3): Accelrys mol/sdf format (V3000).
            - mol2: Tripos Mol2 format.
            - smi: SMILES format. log: Gaussian98/03 output log format. (experimental)
            - pdb: Protein Data Bank format (experimental code)
            - vbf: Visipoint Binary Format (experimental)
            - xyz: XMol XYZ format (experimental code)
        testrun (bool): GA: Test the generated conformer ensemble against the input structure(s). The input
            geometries are used as reference for RMSD calculation, and ensemble statistics are calculated.
            Confomers are generated as if the input was 2D (connectivity only; see option 'rebuildGeometry').
            Typically, ligands from X-ray crystal structures are used as input for a testrun.
        noGA (bool): Do not use a genetic algorithm (GA) to generate conformations. Conformations will be
            generated via distance geometry.
        addConformerNumberToName (bool): Add the number of a conformer as a suffix to its name.
            The suffix is '_X', where X is the number. The name of the first conformer is not altered.
        listSymmetryClasses (bool): List the symmetry class identifiers for each atom of each input structure.
        singleconf (bool): Output only the lowest-energy conformation, regardless of the population size.
        fullforce (bool): GA: Use the full force field in the post-GA structure optimization. The default is to
            ignore torsion _gradient_ and all of electrostatics.
        nosymmetry (bool): GA: Disable the use of symmetry in calculation of RMSD. Symmetry perception can
            be very costly for large structures, e.g., proteins.
        useRingAtomsForRMSD (bool): Consider ring atoms for RMSD calculation instead of non-hydrogen atoms.
            Note that acyclic structures will not be processed at all.
        rebuildGeometry (bool): GA: Do not use the input geometry as template for conformer generation, but always
            rebuild the geometry. See also the 'testrun' option.
        maxPostprocessIter (int): GA: Maximum allowed number of iterations for conjugate gradient structure
            optimization in the post-processing phase. Default = 100.
        maxShapeIterations (int): GA: Maximum allowed number of iterations for overlap optimization in
            shape-matching. Default = 300. The truncated Newton optimizer is adapted from the TNPACK code of
            Schlick and co-workers: Schlick T and Fogelson A (1992) ACM TOMS 18, 46-70 & 71-111; Xie D and
            Schlick T (1999) SIAM J. Opt. 10, 132-154; Xie D and Schlick T (1999) ACM TOMS 25, 108-122.
        maxtime (int): GA: Maximum time [s] used per compound to evolve the GA. Default = 120.
        ftol (float): Tolerance for the change in the objective function value for terminating conjugate
            gradient structure optimization. Pass a negative value to omit this convergence
            criterion. Default is to omit.
        gradientTolerance (float): Tolerance for the gradient root-mean-square (RMS) value for terminating conjugate
            gradient structure optimization. Pass a negative value to omit this convergence criterion.
            Default = 0.050000000000000003.
        nicheRadius (foat): GA: Interconformer distance limit (RMSD) for the calculation of niche count.
            Default = 1.5
        RMSDtol (float): GA: Interconformer distance limit (RMSD) for the final pruning of conformers. If two
            conformers are closer than RMSDtol, the higher-energy conformer will be discarded. Default = 0.5
        energyConstant (int): GA: Constant term [kcal/mol] for the linear function of rotatable bonds used to
            calculate the potential energy window within which the final conformers must reside. Must be >= 0.
            Default = 10.
        energySlope (float): GA: Slope [kcal/mol/rotatable bond] for the linear function of rotatable bonds used to
            calculate the potential energy window within which the final conformers must reside. Must
            be >= 0. Default = 0.5.
        nGenerations (int): GA: The maximum number of generations. Default = 20.
        tournamentSize (int): GA: The number of individuals to involve in the tournament selection. The higher the
            number, the stronger the evolutionary pressure towards geometrically dissimilar results. Default = 2
        pTorsionMutation (float): GA: Mutation probability for changing a torsion angle value.
            Default = 0.050000000000000003
        pStereoMutation (float): GA: Mutation probability for inverting a stereochemical center.
            Default = 0.050000000000000003.
        pPyramidMutation (float): GA: Mutation probability for inverting a pyramidal center.
            Default = 0.050000000000000003
        pRingFlipMutation (float): GA: Mutation probability for a ring flip. Default = 0.050000000000000003
        parallelityThreshold (int): GA: Parallelity threshold [deg] for a ring flip to take effect.
            Valid values within the half-open interval [0,180). Default = 30
        pUniXO (float): GA: Uniform cross-over: probability for crossing. Tested at each locus.
            Default = 0.20000000000000001
        allowcrowded (bool): GA: Do not remove crowded solutions from the population during the evolutionary cycle.
        noPopulationGrowth (bool): GA: Do not allow the population to grow in order to accommodate the Pareto front.
        evolveOverlap (bool):  GA: Evolve the shape-density overlap with a template structure together with the
            conformational energy. Requires a template structure to be specified.
        adjusthydrogens (bool): Add hydrogens to the structures according to the octet rule.
            Hydrogens are always added to structures parsed from SMILES.
        neutralize (bool): Try to add or remove hydrogen atoms in order to make the input molecule neutral.
        stripSalts (bool): Remove all but the largest (highest number of atoms) connected component of the input
            structure.
        expand (bool): Distance geometry: Create expanded conformers. On each iteration, update the
            lower bounds of nonbonded atoms from the previously generated conformer.
        contract (bool): Distance geometry: Create contracted conformers. On each iteration, update the
            upper bounds of nonbonded atoms.
        stereo (bool): Always sample stereoconfigurations. Ignores any stereochemistry specifications in the input.
        keepInitial (bool): Output the initial geometry into the generated set of conformers.
        maxiter (int): Maximum allowed number of iterations for conjugate gradient structure optimization in
            the template geometry generation. Default = 100.
        useSimplex (bool): Use simplex downhill structure optimization prior to conjugate gradient.
        maxSimplexIterations (int): Maximum allowed number of iterations for downhill simplex structure
            optimization. Default = 500.
        simplexStepLength (int): The simplex edge length for downhill simplex structure optimization. Default = 1
        simplexFunctionTolerance (int): Tolerance for the relative change in the objective function value for
            terminating downhill simplex structure optimization. Default = 0.10000000000000001
        nInitialDimensions (int): The number of dimension in which to perform the initial geometry optimization
            after distance geometry. Increase if you encounter incorrect stereochemistry. Default = 4.
        forcefield (str): A filename to read for MMFF94 force field parameters. Alternatively, you can set
            environment variable named BALLOON_FORCEFIELD to point to the parameters file. The command
            line option overrides the environment variable.
        noVdWcutoff (bool): Do not use a cutoff distance for van der Waals (steric) energy evaluation.
        vdWCutoffOn (float): Distance at which the smoothing function for van der Waals energy cutoff is turned on.
        vdWCutoffOff (float): Distance at which the smoothing function for van der Waals energy cutoff is
            turned off.
        noEcutoff (bool): Do not use a cutoff distance for electrostatic energy evaluation.
        ECutoffOn (float): Distance at which the smoothing function for electrostatic energy cutoff is turned on.
        ECutoffOff (float): Distance at which the smoothing function for electrostatic energy cutoff is turned off.
        chargemodel (str): Specify the partial atomic charge model to be used.
            Alternatives are:
            - EEM: Puranen et al. (2010) J. Comput. Chem. 31, 1722-1732. doi:10.1002/jcc.21460
            - MMFF94: Halgren TA (1996) J. Comput. Chem. 17, 490-519.
            - SFKEEM: Puranen et al. (2010) J. Comput. Chem. 31, 1722-1732. doi:10.1002/jcc.21460
        distanceDependent (bool): Use a distance dependent dielectric model in optimizing the initial geometry.
        dielectric (float): Value of the dielectric constant aka relative static permittivity used in the
            calculation of electrostatic energy according to Coulomb's equation. Value must be > 1e-6 in
            order to avoid division by zero. For water, the value is 80 at room temperature. Defaults to one
            for vacuum.
        listAtomTypes (bool): Output a list of assigned force field atom types per each atom.
        maxRingSize (int): Maximum size for rings (#atoms) whose flexibility is to be handled. Negative values
            impose no limit (default).
        maxFlipDistance (int): A maximum allowed number of bonds between the pairs of ring atoms that define a
            flip-of-fragment operation. Negative values impose no limit. Default = 20
        query (str): GA: File name for a query structure upon which the compounds are overlaid based on
            shape-density overlap.
        noElitism (bool): Do not guarantee that the minimum energy conformer always survives to the next generation.
    """

    def __init__(self,
                 balloon_executable_path="",
                 config: str = None,
                 writeMOL2: bool = None,
                 onlycharge: bool = None,
                 nobadmodels: bool = None,
                 strict: bool = None,
                 nconfs: int = None,
                 randomSeed: int = None,
                 sdfnamefieldheader: str = None,
                 output_file: str = None,
                 output_format: str = None,
                 input_file: str = None,
                 testrun: bool = None,
                 noGA: bool = None,
                 addConformerNumberToName: bool = None,
                 listSymmetryClasses: bool = None,
                 singleconf: bool = None,
                 fullforce: bool = None,
                 nosymmetry: bool = None,
                 useRingAtomsForRMSD: bool = None,
                 rebuildGeometry: bool = None,
                 maxPostprocessIter: int = None,
                 maxShapeIterations: int = None,
                 maxtime: int = None,
                 ftol: float = None,
                 gradientTolerance: float = None,
                 nicheRadius: float = None,
                 RMSDtol: float = None,
                 energyConstant: int = None,
                 energySlope: float = None,
                 nGenerations: int = None,
                 tournamentSize: int = None,
                 pTorsionMutation: float = None,
                 pStereoMutation: float = None,
                 pPyramidMutation: float = None,
                 pRingFlipMutation: float = None,
                 parallelityThreshold: int = None,
                 pUniXO: float = None,
                 allowcrowded: bool = None,
                 noPopulationGrowth: bool = None,
                 evolveOverlap: bool = None,
                 adjusthydrogens: bool = None,
                 neutralize: bool = None,
                 stripSalts: bool = None,
                 expand: bool = None,
                 contract: bool = None,
                 stereo: bool = None,
                 keepInitial: bool = None,
                 maxiter: int = None,
                 useSimplex: bool = None,
                 maxSimplexIterations: int = None,
                 simplexStepLength: int = None,
                 simplexFunctionTolerance: int = None,
                 nInitialDimensions: int = None,
                 forcefield: str = None,
                 noVdWcutoff: bool = None,
                 vdWCutoffOn: float = None,
                 vdWCutoffOff: float = None,
                 noEcutoff: bool = None,
                 ECutoffOn: float = None,
                 ECutoffOff: float = None,
                 chargemodel: str = None,
                 distanceDependent: bool = None,
                 dielectric: float = None,
                 listAtomTypes: bool = None,
                 maxRingSize: int = None,
                 maxFlipDistance: int = None,
                 query: str = None,
                 noElitism: bool = None
                 ):
        """Initialize. Pass to config dictionary."""
        self.balloon_executable_path = balloon_executable_path
        self.input_file = input_file
        if self.input_file is not None:
            print("Definition of input file in config is not used. Please pass to run method.")
        self.output_format = output_format
        self.output_file = output_file

        self._config_args = {
            "config": config, "nconfs": nconfs, "randomSeed": randomSeed,
            "sdfnamefieldheader": sdfnamefieldheader, "maxPostprocessIter": maxPostprocessIter,
            "maxShapeIterations": maxShapeIterations, "maxtime": maxtime, "ftol": ftol,
            "gradientTolerance": gradientTolerance, "nicheRadius": nicheRadius, "RMSDtol": RMSDtol,
            "energyConstant": energyConstant, "energySlope": energySlope, "nGenerations": nGenerations,
            "tournamentSize": tournamentSize, "pTorsionMutation": pTorsionMutation, "pStereoMutation": pStereoMutation,
            "pPyramidMutation": pPyramidMutation, "pRingFlipMutation": pRingFlipMutation,
            "parallelityThreshold": parallelityThreshold, "pUniXO": pUniXO, "maxiter": maxiter,
            "maxSimplexIterations": maxSimplexIterations, "simplexStepLength": simplexStepLength,
            "simplexFunctionTolerance": simplexFunctionTolerance, "nInitialDimensions": nInitialDimensions,
            "forcefield": forcefield, "vdWCutoffOn": vdWCutoffOn, "vdWCutoffOff": vdWCutoffOff,
            "ECutoffOn": ECutoffOn, "ECutoffOff": ECutoffOff, "chargemodel": chargemodel, "dielectric": dielectric,
            "maxRingSize": maxRingSize, "maxFlipDistance": maxFlipDistance, "query": query
        }
        self._config_flags = {
            "writeMOL2": writeMOL2, "onlycharge": onlycharge, "nobadmodels": nobadmodels, "strict": strict,
            "testrun": testrun, "noGA": noGA, "addConformerNumberToName": addConformerNumberToName,
            "listSymmetryClasses": listSymmetryClasses, "singleconf": singleconf, "fullforce": fullforce,
            "nosymmetry": nosymmetry, "useRingAtomsForRMSD": useRingAtomsForRMSD, "rebuildGeometry": rebuildGeometry,
            "allowcrowded": allowcrowded, "noPopulationGrowth": noPopulationGrowth, "evolveOverlap": evolveOverlap,
            "adjusthydrogens": adjusthydrogens, "neutralize": neutralize, "stripSalts": stripSalts, "expand": expand,
            "contract": contract, "stereo": stereo, "keepInitial": keepInitial, "useSimplex": useSimplex,
            "noVdWcutoff": noVdWcutoff, "noEcutoff": noEcutoff, "distanceDependent": distanceDependent,
            "listAtomTypes": listAtomTypes, "noElitism": noElitism
        }

    def run(self, input_file: str, output_file: str = None, output_format: str = None):
        if input_file is None:
            raise ValueError("Input file must be defined.")
        if output_file is None:
            output_file = self.output_file
        if output_format is None:
            output_format = self.output_format

        command_list = [str(os.path.join(self.balloon_executable_path, "balloon"))]
        # Added flags. They have the same name as command line argument.
        for key, value in self._config_flags.items():
            if value is not None:
                command_list.append("--"+key)
        # Add args.
        for key, value in self._config_flags.items():
            if value is not None:
                command_list.append("--"+key)
                command_list.append(str(value))
        # Output format
        if output_format is not None:
            command_list.append("--output-format")
            command_list.append(str(output_format))

        # Finally add input and output file
        command_list.append("--input-file")
        command_list.append(str(input_file))

        if output_file is not None:
            command_list.append("--output-file")
            command_list.append(str(output_file))

        return_code = subprocess.run(command_list)
        # Check return code
        if int(return_code.returncode) != 0:
            raise ValueError("Batch process returned with error:", return_code)

        return return_code

    def get_config(self):
        config = {"input_file": self.input_file, "output_format": self.output_format, "output_file": self.output_file,
                  "balloon_executable_path": self.balloon_executable_path}
        config.update(self._config_args)
        config.update(self._config_flags)
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)



import logging

# Module logger
logging.basicConfig()
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


class OneHotEncoder:
    r"""Simple One-Hot-Encoding for python lists. Uses a list of possible values for a one-hot encoding of a single
    value. The translated values must support :obj:``__eq__`` operator.
    The list of possible values must be set beforehand. Is used as a basic encoder example for
    :obj:``MolecularGraphRDKit``. There can not be different dtypes in categories.
    """
    _dtype_translate = {"int": int, "float": float, "str": str, "bool": bool}

    def __init__(self, categories: list, add_unknown: bool = True, dtype: str = "int"):
        """Initialize the encoder beforehand with a set of all possible values to encounter.

        Args:
            categories (list): List of possible values, matching the one-hot encoding.
            add_unknown (bool): Whether to add a unknown bit. Default is True.
            dtype (str): Data type to cast value into before comparing to category entries. Default is "int".
        """
        assert isinstance(dtype, str)
        if dtype not in list(self._dtype_translate.keys()):
            raise ValueError("Unsupported dtype for OneHotEncoder %s" % dtype)
        self.dtype_identifier = dtype
        self.dtype = self._dtype_translate[dtype]
        self.categories = [self.dtype(x) for x in categories]
        self.found_values = []
        self.add_unknown = add_unknown

    def __call__(self, value):
        r"""Encode a single feature or value, mapping it to a one-hot python list. E.g. `[0, 0, 1, 0]`

        Args:
            value: Any object that can be compared to items in ``self.one_hot_values``.

        Returns:
            list: Python List with 1 at value match. E.g. `[0, 0, 1, 0]`
        """
        encoded_list = [1 if x == self.dtype(value) else 0 for x in self.categories]
        if self.add_unknown:
            if value not in self.categories:
                encoded_list += [1]
            else:
                encoded_list += [0]
        if value not in self.found_values:
            self.found_values += [value]
        return encoded_list

    def get_config(self):
        config = {"categories": self.categories, "add_unknown": self.add_unknown, "dtype": self.dtype_identifier}
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def report(self, name=""):
        module_logger.info("OneHotEncoder %s found %s" % (name, self.found_values))


import unittest

import os
import random
import itertools
import tempfile

import numpy as np
import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.literature.MEGAN import MEGAN


class TestMegan(unittest.TestCase):

    def random_input(self, num_batches, num_features):
        n = []
        ei = []
        e = []
        for b in range(num_batches):
            N = random.randint(5, 30)
            node_indices = list(range(N))
            M = random.randint(1, N - 1)

            n.append([[random.random() for _ in range(num_features)] for _ in range(N)])
            e.append([[random.random() for _ in range(num_features)] for _ in range(M)])
            ei.append(list(zip(random.sample(node_indices, M), random.sample(node_indices, M))))

        return (
            tf.ragged.constant(n, ragged_rank=1, dtype=tf.float32),
            tf.ragged.constant(e, ragged_rank=1, dtype=tf.float32),
            tf.ragged.constant(ei, ragged_rank=1, dtype=tf.int32),
        )

    # -- UNITTESTS --

    def test_construction_basically_works(self):
        model = MEGAN(units=[1])
        self.assertIsInstance(model, MEGAN)
        self.assertIsInstance(model, ks.models.Model)

    def test_shapes_basically_work(self):
        num_batches = 5
        num_features = 3
        n, e, ei = self.random_input(num_batches=num_batches, num_features=num_features)

        # We just test a bunch of different configurations for the output dimension and the number of
        # explanation channels.
        output_dimensions = [1, 3]
        channel_dimensions = [2, 6]

        for num_out, num_channels in itertools.product(output_dimensions, channel_dimensions):
            model = MEGAN(
                units=[3],
                importance_channels=num_channels,
                final_units=[num_out],
            )

            out, node_importances, edge_importances = model([n, e, ei])
            self.assertEqual((num_batches, num_out), out.shape)
            self.assertEqual((num_batches, None, num_channels), node_importances.shape)
            self.assertEqual((num_batches, None, num_channels), node_importances.shape)

    def test_saving_loading_basically_works(self):
        num_batches = 5
        num_features = 3
        n, e, eid = self.random_input(num_batches=num_batches, num_features=num_features)

        num_channels = 2
        num_out = 1
        model = MEGAN(
            units=[5, 3],
            importance_channels=num_channels,
            final_units=[num_out]
        )
        # At this point the model is not built yet and should raise a value error
        with self.assertRaises(ValueError):
            model.summary()

        out, ni, ei = model([n, e, eid], training=False)
        self.assertEqual((num_batches, num_out), out.shape)
        self.assertEqual((num_batches, None, num_channels), ni.shape)
        self.assertEqual((num_batches, None, num_channels), ei.shape)

        # After having passed in some input, the model should be built and that should work
        model.summary()
        weights = model.get_weights()

        with tempfile.TemporaryDirectory() as path:
            model_path = os.path.join(path, 'model')

            # Saving the model to a file
            model.save(model_path)
            self.assertTrue(os.path.exists(model_path))

            # Loading the model from that file again
            model_loaded = ks.models.load_model(
                model_path,
                custom_objects={'MEGAN': MEGAN}
            )
            self.assertIsInstance(model_loaded, ks.models.Model)

            out_l, ni_l, ei_l = model_loaded([n, e, eid], training=False)
            # Basic test for correct shape
            self.assertEqual((num_batches, num_out), out_l.shape)
            self.assertEqual((num_batches, None, num_channels), ni_l.shape)
            self.assertEqual((num_batches, None, num_channels), ei_l.shape)

            # Comparing the model weights
            weights_l = model_loaded.get_weights()
            for w, w_l in zip(weights, weights_l):
                np.testing.assert_allclose(w, w_l)

            # Testing if both models produce the same outputs, given the same inputs
            np.testing.assert_allclose(out, out_l, rtol=1e-3)
            np.testing.assert_allclose(ni.values, ni_l.values, rtol=1e-3)
            np.testing.assert_allclose(ei.values, ei_l.values, rtol=1e-3)

    def test_explanation_only_training_basically_works(self):
        num_batches = 20
        num_features = 3
        n, e, eid = self.random_input(num_batches=num_batches, num_features=num_features)
        out = tf.ragged.constant([random.random() for _ in range(num_batches)])

        num_channels = 2
        num_out = 1
        model = MEGAN(
            units=[5, 3],
            importance_channels=num_channels,
            final_units=[num_out],
            return_importances=False,
            # We explicitly want to try using the explanation step here
            importance_factor=1.0,
            # We need this to indicate that we are trying to do regression here:
            regression_limits=(-1, 1),
            regression_reference=0
        )

        model.compile(
            optimizer='adam',
            loss=ks.losses.mean_squared_error
        )
        history = model.fit(
            [n, e, eid], out,
            batch_size=1,
            epochs=2,
            verbose=0
        )
        self.assertIn('exp_loss', history.history)
        self.assertNotEqual(0.0, history.history['exp_loss'])




import numpy as np

from kgcnn.data.datasets.GraphTUDataset2020 import GraphTUDataset2020


class MUTAGDataset(GraphTUDataset2020):
    """Store and process MUTAG dataset."""

    def __init__(self, reload=False, verbose=1):
        r"""Initialize MUTAG dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 0=silent. Default is 1.
        """
        # Use default base class init()
        super(MUTAGDataset, self).__init__("MUTAG", reload=reload, verbose=verbose)

    def read_in_memory(self, verbose: int = 1):
        r"""Load MUTAG data into memory and already split into items."""
        super(MUTAGDataset, self).read_in_memory()

        edge_labels = self.obtain_property("edge_labels")
        node_labels = self.obtain_property("node_labels")
        graph_labels = self.obtain_property("graph_labels")
        # split into separate graphs
        # graph_id, counts = np.unique(mutag_gi, return_counts=True)
        # graphlen = np.zeros(n_data, dtype=np.int)
        # graphlen[graph_id] = counts
        # nodes0123 = np.split(mutag_n, np.cumsum(graphlen)[:-1])
        node_translate = np.array([6, 7, 8, 9, 53, 17, 35], dtype="int")
        atoms_translate = ['C', 'N', 'O', 'F', 'I', 'Cl', 'Br']
        node_attributes = [node_translate[np.array(x, dtype="int")][:, 0] for x in node_labels]
        # nodes = [node_translate[x] for x in nodes0123]
        atoms = [[atoms_translate[int(y)] for y in x] for x in node_labels]
        graph_labels = np.array(graph_labels)
        graph_labels[graph_labels < 0] = 0

        self.assign_property("node_attributes", node_attributes)
        self.assign_property("edge_attributes", [x[:, 0] for x in edge_labels])
        self.assign_property("node_symbol", atoms)
        self.assign_property("node_number", node_attributes)
        self.assign_property("graph_labels", [x for x in graph_labels])
        self.assign_property("graph_size", [len(x) for x in node_attributes])

        return self

# data = MUTAGDataset()



import os
import numpy as np
from collections import defaultdict
from typing import Dict, Callable, List, Union
import pandas as pd
import pymatgen.io.cif
import pymatgen.core.structure
import pymatgen.symmetry.structure

from kgcnn.data.base import MemoryGraphDataset
from kgcnn.data.utils import save_json_file, load_json_file
from kgcnn.crystal.base import CrystalPreprocessor
from kgcnn.graph.base import GraphDict


class CrystalDataset(MemoryGraphDataset):
    r"""Class for making graph dataset from periodic structures such as crystals.

    The dataset class requires a :obj:`data_directory` to store a table '.csv' file containing labels and information
    of the structures stored in either a single '.cif' file or multiple CIF-files in :obj:`file_directory`.
    In the latter case, the file names must be included in the '.csv' table.

    .. code-block:: type

        ├── data_directory
            ├── file_directory
            │   ├── *.cif
            │   ├── *.cif
            │   └── ...
            ├── file_name.csv
            └── file_name.pymatgen.json

    This class uses :obj:`pymatgen.core.structure.Structure` and therefore requires :obj:`pymatgen` to be installed.
    A '.pymatgen.json' serialized file is generated to store a list of structures from '.cif' file format via
    :obj:`prepare_data()`.

    """

    DEFAULT_LOOP_UPDATE_INFO = 5000

    def __init__(self,
                 data_directory: str = None,
                 dataset_name: str = None,
                 file_name: str = None,
                 file_directory: str = None,
                 verbose: int = 10):
        r"""Initialize a base class of :obj:`CrystalDataset`.

        Args:
            data_directory (str): Full path to directory of the dataset. Default is None.
            file_name (str): Filename for dataset to read into memory. This can be a single a 'cif' file.
                Or a '.csv' of file names that are expected to be cif-files in file_directory.
                Default is None.
            file_directory (str): Name or relative path from :obj:`data_directory` to a directory containing sorted
                'cif' files. Default is None.
            dataset_name (str): Name of the dataset. Important for naming and saving files. Default is None.
            verbose (int): Logging level. Default is 10.
        """
        super(CrystalDataset, self).__init__(data_directory=data_directory, dataset_name=dataset_name,
                                             file_name=file_name, verbose=verbose,
                                             file_directory=file_directory)
        self._structs = None

    def _get_pymatgen_file_name(self):
        """Try to determine a file name for the pymatgen serialization information to store to disk."""
        return os.path.splitext(self.file_name)[0] + ".pymatgen.json"

    @staticmethod
    def _pymatgen_serialize_structs(structs: List):
        dicts = []
        for s in structs:
            d = s.as_dict()
            # Module information should be already obtained from as_dict().
            # d["@module"] = type(s).__module__
            # d["@class"] = type(s).__name__
            dicts.append(d)
        return dicts

    @staticmethod
    def _pymatgen_parse_cif_file_to_structures(cif_file: str):
        # structure = pymatgen.io.cif.CifParser.from_string(cif_string).get_structures()[0]
        structures = pymatgen.io.cif.CifParser(cif_file).get_structures()
        return structures

    def prepare_data(self, cif_column_name: str = None, overwrite: bool = False):
        r"""Try to load all crystal structures from CIF files and save them as a pymatgen json serialization.
        Can load a single CIF file with multiple structures (maybe unstable), or multiple CIF files from a table
        that keeps file names and possible labels or additional information.

        Args:
            cif_column_name (str): Name of the column that has file names found in file_directory. Default is None.
            overwrite (bool): Whether to rerun the data extraction. Default is False.

        Returns:
            self
        """
        if os.path.exists(os.path.join(self.data_directory, self._get_pymatgen_file_name())) and not overwrite:
            self.info("Pickled pymatgen structures already exist. Do nothing.")
            return self
        pymatgen_file_made = False

        file_path = os.path.join(self.data_directory, self.file_name)
        file_path_base = os.path.splitext(file_path)[0]

        # Check for a single CIF file.
        found_cif_file = False
        if os.path.exists(file_path_base + ".cif"):
            found_cif_file = True
            self.info("Start to read many structures form cif-file via pymatgen ...")
            structs = self._pymatgen_parse_cif_file_to_structures(file_path)
            self.info("Exporting as dict for pymatgen ...")
            dicts = self._pymatgen_serialize_structs(structs)
            self.info("Saving structures as .json ...")
            out_path = os.path.join(self.data_directory, self._get_pymatgen_file_name())
            save_json_file(dicts, out_path)
            pymatgen_file_made = True

        # We try to read in a csv file.
        self.read_in_table_file(file_path=file_path)

        # Check if table has a list of single cif files in file directory.
        if not found_cif_file and cif_column_name is not None and self.data_frame is not None:
            # Try to find file names in data_frame
            cif_file_list = self.data_frame[cif_column_name].values
            num_structs = len(cif_file_list)
            structs = []
            self.info("Read %s cif-file via pymatgen ..." % num_structs)
            for i, x in enumerate(cif_file_list):
                # Only one file per path
                structs.append(self._pymatgen_parse_cif_file_to_structures(os.path.join(self.data_directory,
                                                                                        self.file_directory, x))[0])
                if i % self.DEFAULT_LOOP_UPDATE_INFO == 0:
                    self.info(" ... read structure {0} from {1}".format(i, num_structs))
            self.info("Exporting as dict for pymatgen ...")
            dicts = self._pymatgen_serialize_structs(structs)
            self.info("Saving structures as .json ...")
            out_path = os.path.join(self.data_directory, self._get_pymatgen_file_name())
            save_json_file(dicts, out_path)
            pymatgen_file_made = True

        if not pymatgen_file_made:
            raise FileNotFoundError("Could not make pymatgen structures.")

        return self

    @staticmethod
    def _pymatgen_deserialize_dicts(dicts: List[dict], to_unit_cell: bool = False) -> list:
        structs = []
        for x in dicts:
            # We could check symmetry or @module, @class items in dict.
            s = pymatgen.core.structure.Structure.from_dict(x)
            structs.append(s)
            if to_unit_cell:
                for site in s.sites:
                    site.to_unit_cell(in_place=True)
        return structs

    def _read_pymatgen_json_in_memory(self):
        file_path = os.path.join(self.data_directory, self._get_pymatgen_file_name())

        if not os.path.exists(file_path):
            raise FileNotFoundError("Cannot find .json file for `CrystalDataset`. Please `prepare_data()`.")

        self.info("Reading structures from .json ...")

        dicts = load_json_file(file_path)
        return self._pymatgen_deserialize_dicts(dicts)

    def _map_callbacks(self,
                       structs: list,
                       data: pd.Series,
                       callbacks: Dict[
                           str, Callable[[pymatgen.core.structure.Structure, pd.Series], Union[np.ndarray, None]]],
                       assign_to_self: bool = True) -> dict:

        # The dictionaries values are lists, one for each attribute defines in "callbacks" and each value in those
        # lists corresponds to one structure in the dataset.
        value_lists = defaultdict(list)
        for index, st in enumerate(structs):
            for name, callback in callbacks.items():
                if st is None:
                    value_lists[name].append(None)
                else:
                    data_dict = data.loc[index]
                    value = callback(st, data_dict)
                    value_lists[name].append(value)
            if index % self.DEFAULT_LOOP_UPDATE_INFO == 0:
                self.info(" ... read structures {0} from {1}".format(index, len(structs)))

        # The string key names of the original "callbacks" dict are also used as the names of the properties which are
        # assigned
        if assign_to_self:
            for name, values in value_lists.items():
                self.assign_property(name, values)

        return value_lists

    def read_in_memory(self, label_column_name: str = None,
                       additional_callbacks: Dict[
                           str, Callable[[pymatgen.core.structure.Structure, pd.Series], None]] = None
                       ):
        """Read structures from pymatgen json serialization and convert them into graph information.

        Args:
            label_column_name (str): Columns of labels for graph. Default is None.
            additional_callbacks (dict): Callbacks to add during read into memory.

        Returns:
            self
        """
        if additional_callbacks is None:
            additional_callbacks = {}

        self.info("Making node features from structure...")
        callbacks = {"graph_labels": lambda st, ds: ds[label_column_name] if label_column_name is not None else None,
                     "node_coordinates": lambda st, ds: np.array(st.cart_coords, dtype="float"),
                     "node_frac_coordinates": lambda st, ds: np.array(st.frac_coords, dtype="float"),
                     "graph_lattice": lambda st, ds: np.ascontiguousarray(np.array(st.lattice.matrix), dtype="float"),
                     "abc": lambda st, ds: np.array(st.lattice.abc),
                     "charge": lambda st, ds: np.array([st.charge], dtype="float"),
                     "volume": lambda st, ds: np.array([st.lattice.volume], dtype="float"),
                     "node_number": lambda st, ds: np.array(st.atomic_numbers, dtype="int"),
                     **additional_callbacks
                     }

        self._map_callbacks(structs=self._read_pymatgen_json_in_memory(),
                            data=self.read_in_table_file(file_path=self.file_path).data_frame,
                            callbacks=callbacks)

        return self

    def set_representation(self, pre_processor: CrystalPreprocessor, reset_graphs: bool = False):

        if reset_graphs:
            self.clear()
        # Read pymatgen JSON file from file.
        structs = self._read_pymatgen_json_in_memory()
        if reset_graphs:
            self.empty(len(structs))

        for index, s in enumerate(structs):
            g = pre_processor(s)
            graph_dict = GraphDict()
            graph_dict.from_networkx(g, node_attributes=pre_processor.node_attributes,
                                     edge_attributes=pre_processor.edge_attributes)
            # TODO: Add graph attributes (label, lattice_matrix, etc.)
            # TODO: Rename node and edge properties to match kgcnn conventions
            # TODO: Add GraphDict to dataset

            if index % self.DEFAULT_LOOP_UPDATE_INFO == 0:
                self.info(" ... preprocess structures {0} from {1}".format(index, len(structs)))

        return self



from sklearn.linear_model import Ridge
import matplotlib.pyplot as plt
import numpy as np

from kgcnn.scaler.scaler import StandardScaler


class ExtensiveMolecularScaler:
    """Scaler for extensive properties like energy to remove a simple linear behaviour with additive atom
    contributions. Interface is designed after scikit-learn standard scaler. Internally Ridge regression ist used.
    Only the atomic number is used as extensive scaler. This could be further improved by also taking bonds and
    interactions into account, e.g. as energy contribution.

    """
    _attributes_list_sklearn = ["coef_", "intercept_", "n_iter_", "n_features_in_", "feature_names_in_"]
    _attributes_list_mol = ["scale_", "_fit_atom_selection", "_fit_atom_selection_mask"]
    max_atomic_number = 95

    def __init__(self, alpha: float = 1e-9, fit_intercept: bool = False, **kwargs):
        r"""Initialize scaler with parameters directly passed to scikit-learns :obj:`Ridge()`.

        Args:
            alpha (float): Regularization parameter for regression.
            fit_intercept (bool): Whether to allow a constant offset per target.
            kwargs: Additional arguments passed to :obj:`Ridge()`.
        """

        self.ridge = Ridge(alpha=alpha, fit_intercept=fit_intercept, **kwargs)

        self._fit_atom_selection_mask = None
        self._fit_atom_selection = None
        self.scale_ = None

    def fit(self, atomic_number, molecular_property, sample_weight=None):
        r"""Fit atomic number to the molecular properties.

        Args:
            atomic_number (list): List of array of atomic numbers. Shape is `(n_samples, #atoms)`.
            molecular_property (np.ndarray): Array of atomic properties of shape `(n_samples, n_properties)`.
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:
            self
        """
        if len(atomic_number) != len(molecular_property):
            raise ValueError(
                "`ExtensiveMolecularScaler` different input shape {0} vs. {1}".format(
                    len(atomic_number), len(molecular_property))
            )

        unique_number = [np.unique(x, return_counts=True) for x in atomic_number]
        all_unique = np.unique(np.concatenate([x[0] for x in unique_number], axis=0))
        self._fit_atom_selection = all_unique
        atom_mask = np.zeros(self.max_atomic_number, dtype="bool")
        atom_mask[all_unique] = True
        self._fit_atom_selection_mask = atom_mask
        total_number = []
        for unique_per_mol, num_unique in unique_number:
            array_atoms = np.zeros(self.max_atomic_number)
            array_atoms[unique_per_mol] = num_unique
            positives = array_atoms[atom_mask]
            total_number.append(positives)
        total_number = np.array(total_number)
        self.ridge.fit(total_number, molecular_property, sample_weight=sample_weight)
        diff = molecular_property - self.ridge.predict(total_number)
        self.scale_ = np.std(diff, axis=0, keepdims=True)
        return self

    def predict(self, atomic_number):
        """Predict the offset form atomic numbers. Requires :obj:`fit()` called previously.

        Args:
            atomic_number (list): List of array of atomic numbers. Shape is `(n_samples, <#atoms>)`.

        Returns:
            np.ndarray: Offset of atomic properties fitted previously. Shape is `(n_samples, n_properties)`.
        """
        if self._fit_atom_selection_mask is None:
            raise ValueError("ERROR: `ExtensiveMolecularScaler` has not been fitted yet. Can not predict.")
        unique_number = [np.unique(x, return_counts=True) for x in atomic_number]
        total_number = []
        for unique_per_mol, num_unique in unique_number:
            array_atoms = np.zeros(self.max_atomic_number)
            array_atoms[unique_per_mol] = num_unique
            positives = array_atoms[self._fit_atom_selection_mask]
            if np.sum(positives) != np.sum(num_unique):
                print("`ExtensiveMolecularScaler` got unknown atom species in transform.")
            total_number.append(positives)
        total_number = np.array(total_number)
        offset = self.ridge.predict(total_number)
        return offset

    def _plot_predict(self, atomic_number, molecular_property):
        """Debug function to check prediction."""
        molecular_property = np.array(molecular_property)
        if len(molecular_property.shape) <= 1:
            molecular_property = np.expand_dims(molecular_property, axis=-1)
        predict_prop = self.predict(atomic_number)
        if len(predict_prop.shape) <= 1:
            predict_prop = np.expand_dims(predict_prop, axis=-1)
        mae = np.mean(np.abs(molecular_property - predict_prop), axis=0)
        plt.figure()
        for i in range(predict_prop.shape[-1]):
            plt.scatter(predict_prop[:, i], molecular_property[:, i], alpha=0.3,
                        label="Pos: " + str(i) + " MAE: {0:0.4f} ".format(mae[i]))
        plt.plot(np.arange(np.amin(molecular_property), np.amax(molecular_property), 0.05),
                 np.arange(np.amin(molecular_property), np.amax(molecular_property), 0.05), color='red')
        plt.xlabel('Fitted')
        plt.ylabel('Actual')
        plt.legend(loc='upper left', fontsize='x-small')
        plt.show()

    def transform(self, atomic_number, molecular_property):
        """Transform any atomic number list with matching properties based on previous fit. Also std-scaled.

        Args:
            atomic_number (list): List of array of atomic numbers. Shape is `(n_samples, <#atoms>)`.
            molecular_property (np.ndarray): Array of atomic properties of shape `(n_samples, n_properties)`.

        Returns:
            np.ndarray: Transformed atomic properties fitted. Shape is `(n_samples, n_properties)`.
        """
        return (molecular_property - self.predict(atomic_number)) / self.scale_

    def fit_transform(self, atomic_number, molecular_property, sample_weight=None):
        """Combine fit and transform methods in one call.

        Args:
            atomic_number (list): List of array of atomic numbers. Shape is `(n_samples, <#atoms>)`.
            molecular_property (np.ndarray): Array of atomic properties of shape `(n_samples, n_properties)`.
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:
            np.ndarray: Transformed atomic properties fitted. Shape is `(n_samples, n_properties)`.
        """
        self.fit(atomic_number, molecular_property, sample_weight)
        return self.transform(atomic_number, molecular_property)

    def inverse_transform(self, atomic_number, molecular_property):
        """Reverse the transform method to original properties without offset and scaled to original units.

        Args:
            atomic_number (list): List of array of atomic numbers. Shape is `(n_samples, <#atoms>)`.
            molecular_property (np.ndarray): Array of atomic properties of shape `(n_samples, n_properties)`

        Returns:
            np.ndarray: Original atomic properties. Shape is `(n_samples, n_properties)`.
        """
        return molecular_property * self.scale_ + self.predict(atomic_number)

    def get_config(self):
        """Get configuration for scaler."""
        return self.ridge.get_params()

    def get_weights(self) -> dict:
        weights = dict()
        for x in self._attributes_list_mol:
            weights.update({x: np.array(getattr(self, x))})
        for x in self._attributes_list_sklearn:
            if hasattr(self.ridge, x):
                weights.update({x: np.array(getattr(self.ridge, x))})
        return weights


class QMGraphLabelScaler:
    """A scaler that scales QM targets differently. For now, the main difference is that intensive and extensive
    properties are scaled differently. In principle, also dipole, polarizability or rotational constants
    could to be standardized differently.

    """

    def __init__(self, scaler: list):
        if not isinstance(scaler, list):
            raise TypeError("Scaler information for `QMGraphLabelScaler` must be list, got %s." % scaler)

        self.scaler_list = []
        for x in scaler:
            # If x is already a scaler, add it directly to the scaler list.
            if hasattr(x, "fit") and hasattr(x, "transform") and hasattr(x, "inverse_transform"):
                self.scaler_list.append(x)
                continue
            # Otherwise, must be serialized version of a scaler.
            if not isinstance(x, dict):
                raise TypeError("Single scaler for `QMGraphLabelScaler` must be dict, got %s." % x)

            if "class_name" not in x:
                raise ValueError("Scaler class for single target must be defined, got %s" % x)

            if x["class_name"] == "StandardScaler":
                self.scaler_list.append(StandardScaler(**x["config"]))
            elif x["class_name"] == "ExtensiveMolecularScaler":
                self.scaler_list.append(ExtensiveMolecularScaler(**x["config"]))
            else:
                raise ValueError("Unsupported scaler %s" % x["name"])

        self.scale_ = None

    def _input_for_each_scaler_type(self, scaler, graph_labels, node_number):
        if isinstance(scaler, StandardScaler):
            return [graph_labels]
        elif isinstance(scaler, ExtensiveMolecularScaler):
            return node_number, graph_labels
        raise TypeError("Unsupported scaler %s" % scaler)

    def _scale_for_each_scaler_type(self, scaler):
        if isinstance(scaler, StandardScaler):
            return scaler.scale_
        elif isinstance(scaler, ExtensiveMolecularScaler):
            return scaler.scale_[0]
        raise TypeError("Unsupported scaler %s" % scaler)

    def fit_transform(self, graph_labels, node_number):
        r"""Fit and transform all target labels for QM9.

        Args:
            graph_labels (np.ndarray): Array of QM9 labels of shape `(N, 15)`.
            node_number (list): List of atomic numbers for each molecule. E.g. `[np.array([6,1,1,1]), ...]`.

        Returns:
            np.ndarray: Transformed labels of shape `(N, 15)`.
        """
        self.fit(graph_labels, node_number)
        return self.transform(graph_labels, node_number)

    def transform(self, graph_labels, node_number):
        r"""Transform all target labels for QM. Requires :obj:`fit()` called previously.

        Args:
            graph_labels (np.ndarray): Array of QM unscaled labels of shape `(N, #labels)`.
            node_number (list): List of atomic numbers for each molecule. E.g. `[np.array([6,1,1,1]), ...]`.

        Returns:
            np.ndarray: Transformed labels of shape `(N, #labels)`.
        """
        self._check_input(node_number, graph_labels)

        out_labels = []
        for i, x in enumerate(self.scaler_list):
            labels = graph_labels[:, i:i+1]
            out_labels.append(x.transform(*self._input_for_each_scaler_type(x, labels, node_number)))

        out_labels = np.concatenate(out_labels, axis=-1)
        return out_labels

    def fit(self, graph_labels, node_number):
        r"""Fit scaling of QM9 graph labels or targets.

        Args:
            graph_labels (np.ndarray): Array of QM labels of shape `(N, #labels)`.
            node_number (list): List of atomic numbers for each molecule. E.g. `[np.array([6,1,1,1]), ...]`.

        Returns:
            self
        """
        self._check_input(node_number, graph_labels)

        for i, x in enumerate(self.scaler_list):
            labels = graph_labels[:, i:i + 1]
            x.fit(*self._input_for_each_scaler_type(x, labels, node_number))

        self.scale_ = np.concatenate([self._scale_for_each_scaler_type(x) for x in self.scaler_list], axis=0)
        return self

    def inverse_transform(self, graph_labels, node_number):
        r"""Back-transform all target labels for QM9.

        Args:
            graph_labels (np.ndarray): Array of QM scaled labels of shape `(N, #labels)`.
            node_number (list): List of atomic numbers for each molecule. E.g. `[np.array([6,1,1,1]), ...]`.

        Returns:
            np.ndarray: Back-transformed labels of shape `(N, 15)`.
        """
        self._check_input(node_number, graph_labels)

        out_labels = []
        for i, x in enumerate(self.scaler_list):
            labels = graph_labels[:, i:i + 1]
            out_labels.append(x.inverse_transform(*self._input_for_each_scaler_type(x, labels, node_number)))

        out_labels = np.concatenate(out_labels, axis=-1)
        return out_labels

    def _check_input(self, node_number, graph_labels):
        assert len(node_number) == len(graph_labels), "`QMGraphLabelScaler` input length does not match."
        assert graph_labels.shape[-1] == len(self.scaler_list), "`QMGraphLabelScaler` got wrong number of labels."



import numpy as np
import re
import logging
import networkx as nx
from typing import Union
from kgcnn.graph.adapter import GraphTensorMethodsAdapter
from copy import deepcopy

logging.basicConfig()  # Module logger
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


class GraphDict(dict, GraphTensorMethodsAdapter):
    r"""Dictionary container to store graph information in tensor-form. At the moment only numpy-arrays are supported.
    The naming convention is not restricted. The class is supposed to be handled just as a python dictionary.
    In addition, :obj:`assign_property` and :obj:`obtain_property` handles `None` values and cast into tensor format,
    when assigning a named value.

    Graph operations that modify edges or sort indices are methods of this class supported by
    :obj:`kgcnn.graph.adapter.GraphTensorMethodsAdapter`.
    Note that the graph-tensors name must follow a standard-convention or be provided to member functions
    (see documentation of :obj:`kgcnn.graph.adapter.GraphTensorMethodsAdapter`).

    .. code-block:: python

        import numpy as np
        from kgcnn.graph.base import GraphDict
        g = GraphDict({"edge_indices": np.array([[1, 0], [0, 1]]), "edge_labels": np.array([[-1], [1]])})
        g.add_edge_self_loops().sort_edge_indices()  # from GraphTensorMethodsAdapter
        print(g)
    """

    # Implementation details: Inherits from python-dict at the moment but can be changed if this causes problems,
    # alternatives would be: collections.UserDict or collections.abc.MutableMapping

    def __init__(self, sub_dict: dict = None):
        r"""Initialize a new :obj:`GraphDict` instance.

        Args:
            sub_dict: Dictionary or key-value pair of numpy arrays.
        """
        self._tensor_conversion = np.array
        if sub_dict is None:
            sub_dict = {}
        elif isinstance(sub_dict, (dict, list)):
            in_dict = dict(sub_dict)
            sub_dict = {key: self._tensor_conversion(value) for key, value in in_dict.items()}
        elif isinstance(sub_dict, GraphDict):
            sub_dict = {key: self._tensor_conversion(value) for key, value in sub_dict.items()}
        super(GraphDict, self).__init__(sub_dict)

    def to_dict(self) -> dict:
        """Returns a python-dictionary of self. Does not copy values.

        Returns:
            dict: Dictionary of graph tensor objects.
        """
        return {key: value for key, value in self.items()}

    def from_networkx(self, graph,
                      node_number: str = "node_number",
                      edge_indices: str = "edge_indices",
                      node_attributes: str = None,
                      edge_attributes: str = None,
                      node_labels: str = None):
        r"""Convert a networkx graph instance into a dictionary of graph-tensors. The networkx graph is always converted
        into integer node labels. The former node IDs can be hold in :obj:`node_labels`. Furthermore, node or edge
        data can be cast into attributes via :obj:`node_attributes` and :obj:`edge_attributes`.

        Args:
            graph (nx.Graph): A networkx graph instance to convert.
            node_number (str): The name that the node numbers are assigned to. Default is "node_number".
            edge_indices (str): The name that the edge indices are assigned to. Default is "edge_indices".
            node_attributes (str, list): Name of node attributes to add from node data. Can also be a list of names.
                Default is None.
            edge_attributes (str, list): Name of edge attributes to add from edge data. Can also be a list of names.
                Default is None.
            node_labels (str): Name of the labels of nodes to store former node IDs into. Default is None.

        Returns:
            self.
        """
        assert node_labels is None or isinstance(node_labels, str), "Please provide name of node labels or `None`"
        graph_int = nx.convert_node_labels_to_integers(graph, label_attribute=node_labels)
        graph_size = len(graph_int)

        def _attr_to_list(attr):
            if attr is None:
                attr = []
            elif isinstance(attr, str):
                attr = [attr]
            if not isinstance(attr, list):
                raise TypeError("Attribute name is neither list nor string.")
            return attr

        # Loop over nodes in graph.
        node_attr = _attr_to_list(node_attributes)
        if node_labels is not None:
            node_attr += [node_labels]
        node_attr_dict = {x: [None]*graph_size for x in node_attr}
        nodes_id = []
        for i, x in enumerate(graph_int.nodes.data()):
            nodes_id.append(x[0])
            for d in node_attr:
                if d not in x[1]:
                    raise KeyError("Node does not have property %s" % d)
                node_attr_dict[d][i] = x[1][d]

        edge_id = []
        edges_attr = _attr_to_list(edge_attributes)
        edges_attr_dict = {x: [None]*graph.number_of_edges() for x in edges_attr}
        for i, x in enumerate(graph_int.edges.data()):
            edge_id.append(x[:2])
            for d in edges_attr:
                if d not in x[2]:
                    raise KeyError("Edge does not have property %s" % d)
                edges_attr_dict[d][i] = x[2][d]

        # Storing graph tensors in self.
        self.assign_property(node_number, self._tensor_conversion(nodes_id))
        self.assign_property(edge_indices, self._tensor_conversion(edge_id))
        for key, value in node_attr_dict.items():
            self.assign_property(key, self._tensor_conversion(value))
        for key, value in edges_attr_dict.items():
            self.assign_property(key, self._tensor_conversion(value))
        return self

    def to_networkx(self, edge_indices: str = "edge_indices"):
        """Function draft to make a networkx graph. No attributes or data is supported at the moment.

        Args:
            edge_indices (str): Name of edge index tensors to make graph with. Default is "edge_indices".

        Returns:
            nx.DiGraph: Directed networkx graph instance.
        """
        graph = nx.DiGraph()
        graph.add_edges_from(self.obtain_property(edge_indices))
        return graph

    def assign_property(self, key: str, value):
        r"""Add a named property as key, value pair to self. If the value is `None`, nothing is done.
        Similar to assign-item default method :obj:`__setitem__`, but ignores `None` values and casts to tensor.

        Args:
            key (str): Name of the graph tensor to add to self.
            value: Array or tensor value to add. Can also be None.

        Returns:
            None.
        """
        if value is not None:
            self.update({key: self._tensor_conversion(value)})

    def obtain_property(self, key: str):
        """Get tensor item by name. If key is not found, `None` is returned.

        Args:
            key (str): Name of the key to get value for.

        Returns:
            self[key].
        """
        if key in self:
            return self[key]
        return None

    def search_properties(self, keys: Union[str, list]) -> list:
        r"""Search for properties in self.

        This includes a list of possible names or a pattern-matching of a single string.

        Args:
            keys (str, list): Pattern matching string or list of strings to search for.

        Returns:
            list: List of names in self that match :obj:`keys`.
        """
        if keys is None:
            return []
        elif isinstance(keys, str):
            match_props = []
            for x in self:
                if re.match(keys, x):
                    if re.match(keys, x).group() == x:
                        match_props.append(x)
            return sorted(match_props)
        elif isinstance(keys, (list, tuple)):
            # No pattern matching for list input.
            return sorted([x for x in keys if x in self])
        return []
    # Old Alias
    find_graph_properties = search_properties

    def assert_has_valid_key(self, key: str, raise_error: bool = True):
        """Assert the property is found in self.

        Args:
            key (str): Name of property that must be defined.
            raise_error (bool): Whether to raise error. Default is True.

        Returns:
            bool: Key is valid. Only if `raise_error` is False.
        """
        if key not in self or self.obtain_property(key) is None:
            if raise_error:
                raise AssertionError("`GraphDict` does not have %s." % key)
            return False
        return True

    def has_valid_key(self, key: str):
        """Check if the property is found in self and also is not Noe.

        Args:
            key (str): Name of property that must be defined.

        Returns:
            bool: Key is valid. Only if `raise_error` is False.
        """
        if key not in self or self.obtain_property(key) is None:
            return False
        return True

    # Alias of internal assign and obtain property.
    set = assign_property
    # get = obtain_property  # Already has correct behaviour.


GraphNumpyContainer = GraphDict


class GraphPreProcessorBase:

    def __init__(self, *, in_place: bool = True, name: str = None):
        self._config_kwargs = {"in_place": in_place, "name": name}
        self._to_obtain = {}
        self._to_assign = None
        self._call_kwargs = {}
        self._search = []
        self._quite = []
        self._in_place = in_place

    def _obtain_properties(self, graph: GraphDict):
        obtained_properties = {}
        for key, name in self._to_obtain.items():
            if isinstance(name, str):
                if name in self._search:
                    name = graph.search_properties(name)  # Will be sorted list and existing only
                    obtained_properties[key] = [graph.obtain_property(x) for x in name]
                else:
                    if not graph.has_valid_key(name) and key not in self._quite:
                        module_logger.warning("Missing '%s' in '%s'" % (name, type(graph).__name__))
                    obtained_properties[key] = graph.obtain_property(name)
            elif isinstance(name, (list, tuple)):
                prop_list = []
                for x in name:
                    if not graph.has_valid_key(x) and key not in self._quite:
                        module_logger.warning("Missing '%s' in '%s'" % (x, type(graph).__name__))
                    prop_list.append(graph.obtain_property(x))
                obtained_properties[key] = prop_list
            else:
                raise ValueError("Unsupported property identifier %s" % name)

        return obtained_properties

    def _assign_properties(self, graph: GraphDict, graph_properties: Union[list, np.ndarray]):

        def _assign_single(name, single_graph_property):
            if name in self._search:
                names = graph.search_properties(name)  # Will be sorted list and existing only
                # Assume that names matches graph_properties
                if not isinstance(single_graph_property, (list, tuple)):
                    module_logger.error("Wrong return type for %s" % name)
                if len(names) != len(single_graph_property):
                    module_logger.error("Wrong number of %s for %s" % (names, single_graph_property))
                for x, gp in zip(names, single_graph_property):
                    graph.assign_property(x, gp)
                return
            graph.assign_property(name, single_graph_property)
            return

        if isinstance(self._to_assign, str):
            _assign_single(self._to_assign, graph_properties)

        if isinstance(self._to_assign, (list, tuple)):
            if not isinstance(graph_properties, (list, tuple)):
                module_logger.error("Wrong return type for %s" % self._to_assign)
            if len(self._to_assign) != len(graph_properties):
                module_logger.error("Wrong number of %s for %s" % (self._to_assign, graph_properties))
            for key, value in zip(self._to_assign, graph_properties):
                _assign_single(key, value)
            return

    def call(self, **kwargs):
        raise NotImplementedError("Must be implemented in sub-class.")

    def __call__(self, graph: GraphDict):
        if not self._in_place:
            graph = GraphDict(graph)
        graph_properties = self._obtain_properties(graph)
        processed_properties = self.call(**graph_properties, **self._call_kwargs)
        self._assign_properties(graph, processed_properties)
        return graph

    @staticmethod
    def has_special_characters(query, pat=re.compile("[@\\\\!#$%^&*()<>?/|}{~:]")):
        return pat.search(query) is not None

    def get_config(self):
        config = deepcopy(self._config_kwargs)
        return config



import tensorflow as tf

from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.gather import GatherNodesOutgoing, GatherNodesIngoing
from kgcnn.layers.pooling import PoolingLocalEdges
from kgcnn.layers.modules import LazySubtract


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='DMPNNGatherEdgesPairs')
class DMPNNGatherEdgesPairs(GraphBaseLayer):
    """Gather edge pairs that also works for invalid indices given a certain pair, i.e. if a edge does not have its
    reverse counterpart in the edge indices list.
    This class is used in `DMPNN <https://pubs.acs.org/doi/full/10.1021/acs.jcim.9b00237>`_ .

    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(DMPNNGatherEdgesPairs, self).__init__(**kwargs)
        self.gather_layer = GatherNodesIngoing()

    def build(self, input_shape):
        """Build layer."""
        super(DMPNNGatherEdgesPairs, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [edges, pair_index]

                - edges (tf.RaggedTensor): Node embeddings of shape (batch, [M], F)
                - pair_index (tf.RaggedTensor): Edge indices referring to edges of shape (batch, [M], 1)

        Returns:
            list: Gathered edge embeddings that match the reverse edges of shape (batch, [M], F) for selection_index.
        """
        self.assert_ragged_input_rank(inputs)
        edges, pair_index = inputs
        index_corrected = tf.RaggedTensor.from_row_splits(
            tf.where(pair_index.values >= 0, pair_index.values, tf.zeros_like(pair_index.values)),
            pair_index.row_splits, validate=self.ragged_validate)
        edges_paired = self.gather_layer([edges, index_corrected], **kwargs)
        edges_corrected = tf.RaggedTensor.from_row_splits(
            tf.where(pair_index.values >= 0, edges_paired.values, tf.zeros_like(edges_paired.values)),
            edges_paired.row_splits, validate=self.ragged_validate)
        return edges_corrected


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='DMPNNPPoolingEdgesDirected')
class DMPNNPPoolingEdgesDirected(GraphBaseLayer):
    """Pooling of edges for around a target node as defined by
    `DMPNN <https://pubs.acs.org/doi/full/10.1021/acs.jcim.9b00237>`_ . This slightly different than the normal node
    aggregation from message passing like networks. Requires edge pairs for this implementation.

    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(DMPNNPPoolingEdgesDirected, self).__init__(**kwargs)
        self.pool_edge_1 = PoolingLocalEdges(pooling_method="sum")
        self.gather_edges = GatherNodesOutgoing()
        self.gather_pairs = DMPNNGatherEdgesPairs()
        self.subtract_layer = LazySubtract()

    def build(self, input_shape):
        """Build layer."""
        super(DMPNNPPoolingEdgesDirected, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [nodes, edges, edge_index, edge_reverse_pair]

                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
                - edges (tf.RaggedTensor): Edge or message embeddings of shape (batch, [M], F)
                - edge_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
                - edge_reverse_pair (tf.RaggedTensor): Pair mappings for reverse edges (batch, [M], 1)

        Returns:
            tf.RaggedTensor: Edge embeddings of shape (batch, [M], F)
        """
        n, ed, edi, edp = inputs
        pool_edge_receive = self.pool_edge_1([n, ed, edi], **kwargs)  # Sum pooling of all edges
        ed_new = self.gather_edges([pool_edge_receive, edi], **kwargs)
        ed_not = self.gather_pairs([ed, edp],  **kwargs)
        out = self.subtract_layer([ed_new, ed_not],  **kwargs)
        return out



import os
import pickle
import numpy as np
import scipy.io
import json
import pandas as pd

from kgcnn.data.qm import QMDataset
from kgcnn.data.download import DownloadDataset
from kgcnn.mol.io import write_list_to_xyz_file
from kgcnn.graph.geom import coulomb_matrix_to_inverse_distance_proton, coordinates_from_distance_matrix
from kgcnn.graph.adj import invert_distance
from kgcnn.mol.methods import inverse_global_proton_dict


class QM7Dataset(QMDataset, DownloadDataset):
    """Store and process QM7b dataset."""

    download_info = {
        "dataset_name": "QM7",
        "data_directory_name": "qm7",
        # https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm7.mat
        "download_url": "http://quantum-machine.org/data/qm7.mat",
        "download_file_name": 'qm7.mat',
        "unpack_tar": False,
        "unpack_zip": False,
    }

    def __init__(self, reload: bool = False, verbose: int = 1):
        """Initialize QM9 dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 0=silent. Default is 1.
        """
        QMDataset.__init__(self, verbose=verbose, dataset_name="QM7")
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        self.label_names = ["u0_atom"]
        self.label_units = ["kcal/mol"]
        self.label_unit_conversion = np.array([1.0] * 14)  # Pick always same units for training
        self.dataset_name = "QM7"
        self.require_prepare_data = True
        self.fits_in_memory = True
        self.verbose = verbose
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name)
        self.file_name = "qm7.xyz"

        if self.require_prepare_data:
            self.prepare_data(overwrite=reload)

        if self.fits_in_memory:
            self.read_in_memory(label_column_name=self.label_names)

    def prepare_data(self, overwrite: bool = False, xyz_column_name: str = None, make_sdf: bool = True):
        file_path = os.path.join(self.data_directory, self.file_name)
        if not os.path.exists(file_path) or overwrite:
            mat = scipy.io.loadmat(os.path.join(self.data_directory, self.download_info["download_file_name"]))
            graph_len = [int(np.around(np.sum(x > 0))) for x in mat["Z"]]
            proton = [x[:i] for i, x in zip(graph_len, mat["Z"])]
            atoms = [[inverse_global_proton_dict[i] for i in x] for x in proton]
            pos = [x[:i, :]*0.529177210903 for i, x in zip(graph_len, mat["R"])]
            atoms_pos = [[x, y] for x, y in zip(atoms, pos)]
            np.save(os.path.join(self.data_directory, "qm7_splits.npy"), mat["P"])
            self.info("Writing XYZ file from coulomb matrix information.")
            write_list_to_xyz_file(file_path, atoms_pos)
        else:
            self.info("Found XYZ file for qm7b already created.")

        file_path = os.path.join(self.data_directory, os.path.splitext(self.file_name)[0] + ".csv")
        if not os.path.exists(file_path) or overwrite:
            mat = scipy.io.loadmat(os.path.join(self.data_directory, self.download_info["download_file_name"]))
            labels = mat["T"][0]
            targets = pd.DataFrame(labels, columns=self.label_names)
            self.info("Writing CSV file of graph labels.")
            targets.to_csv(file_path, index=False)
        else:
            self.info("Found CSV file of graph labels.")

        return super(QM7Dataset, self).prepare_data(
            overwrite=overwrite, xyz_column_name=xyz_column_name, make_sdf=make_sdf)

    def read_in_memory_sdf(self, **kwargs):
        super(QM7Dataset, self).read_in_memory_sdf()

        # Mean molecular weight mmw
        mass_dict = {'H': 1.0079, 'C': 12.0107, 'N': 14.0067, 'O': 15.9994, 'F': 18.9984, 'S': 32.065}

        def mmw(atoms):
            mass = [mass_dict[x] for x in atoms]
            return np.array([np.mean(mass), len(mass)])

        self.assign_property("graph_attributes", [mmw(x) for x in self.obtain_property("node_symbol")])

# data = QM7Dataset(reload=False)



import logging


def parse_list_to_xyz_str(mol: list, comment: str = ""):
    """Convert list of atom and coordinates list into xyz-string.

    Args:
        mol (list): Tuple or list of `[['C', 'H', ...], [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0], ... ]]`.
        comment (str): Comment for comment line in xyz string. Default is "".

    Returns:
        str: Information in xyz-string format.
    """
    atoms = mol[0]
    coordinates = mol[1]
    if len(atoms) != len(coordinates):
        raise ValueError("Number of atoms does not match number of coordinates for xyz string.")
    xyz_str = str(int(len(atoms))) + "\n"
    if "\n" in comment:
        raise ValueError("Line break must not be in the comment line for xyz string.")
    xyz_str = xyz_str + comment + "\n"
    for a_iter, c_iter in zip(atoms, coordinates):
        _at_str = str(a_iter)
        _c_format_str = " {:.10f}" * len(c_iter) + "\n"
        xyz_str = xyz_str + _at_str + _c_format_str.format(*c_iter)
    return xyz_str


def write_list_to_xyz_file(filepath: str, mol_list: list):
    """Write a list of nested list of atom and coordinates into xyz-string. Uses :obj:`parse_list_to_xyz_str`.

    Args:
        filepath (str): Full path to file including name.
        mol_list (list): List of molecules, which is a list of pairs of atoms and coordinates of
            `[[['C', 'H', ... ], [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0], ... ]], ... ]`.
    """
    with open(filepath, "w+") as file:
        for x in mol_list:
            xyz_str = parse_list_to_xyz_str(x)
            file.write(xyz_str)


def parse_mol_str(mol_str: str):
    """Parse a MDL mol table string into nested list. Only supports V2000 format and CTab. Better rely on
    OpenBabel to do this. This function was a temporary solution.

    Args:
        mol_str (str): String of mol block.

    Returns:
        list: [title, program, comment, counts, atoms, bonds, properties]
    """
    empty_return = ["", "", "", [], [], [], []]
    if len(mol_str) == 0:
        logging.error("Received empty MLD mol-block string. Nothing to parse. Return empty list.")
        return empty_return
    lines = mol_str.split("\n")
    if len(lines) < 4:
        logging.error("Could not find counts line. Invalid format. Can not parse string. Return empty list.")
        return empty_return

    title = lines[0]
    program = lines[1]  # IIPPPPPPPPMMDDYYHHmmddSSssssssssssEEEEEEEEEEEERRRRRR
    comment = lines[2]
    version = lines[3][-6:].strip()
    if version == "V2000":
        # counts has aaabbblllfffcccsssxxxrrrpppiiimmmvvvvvv
        # or shorter but should have version of len=5 at the end
        counts = [lines[3][i:i + 3].strip() for i in range(0, len(lines[3][:-6]), 3)] + [version]
        na = int(counts[0])
        nb = int(counts[1])
        nl = int(counts[2])
        ns = int(counts[5])
        if ns != 0 or nl != 0:
            logging.warning("Not supporting atom lists (deprecated) or stext entries for this function.")
        atoms = []
        for a in lines[4:(na + 4)]:
            # xxxxx.xxxxyyyyy.yyyyzzzzz.zzzz aaaddcccssshhhbbbvvvHHHrrriiimmmnnneee
            # noinspection PyTypeChecker
            atoms.append([a[0:10].strip(), a[10:20].strip(), a[20:30].strip(), a[30:34].strip(), a[34:36].strip(),
                          a[36:39].strip(), a[39:42].strip(), a[42:45].strip(), a[45:48].strip(), a[48:51].strip(),
                          a[51:54].strip(), a[54:57].strip(), a[57:60].strip(), a[60:63].strip(), a[63:66].strip(),
                          a[66:69].strip()])
        # bond block
        bonds = []
        for b in lines[4 + na:4 + na + nb]:
            # 111222tttsssxxxrrrccc
            # noinspection PyTypeChecker
            bonds.append([b[i:i+3].strip() for i in range(0, len(b), 3)])
        # Properties block
        properties = []
        for p in lines[4 + na + nb:]:
            if p == "M  END":
                break
            if "M" in p:
                properties.append(p)
    else:
        raise NotImplementedError("Can not parse mol V3000 or higher.")
    return [title, program, comment, counts, atoms, bonds, properties]


def read_xyz_file(file_path, delimiter: str = None, line_by_line=False):
    """Simple python script to read xyz-file and parse into a nested python list. Always returns a list with
    the geometries in xyz file.

    Args:
        file_path (str): Full path to xyz-file.
        delimiter (str): Delimiter for xyz separation. Default is ' '.
        line_by_line (bool): Whether to read XYZ file line by line.

    Returns:
        list: Nested coordinates from xyz-file.
    """
    mol_list = []
    comment_list = []
    # open file
    infile = open(file_path, "r")
    if line_by_line:
        lines = infile  # File object
    else:
        lines = infile.readlines()  # list of lines

    num = 0
    comment = 0
    atoms = []
    coordinates = []
    for line in lines:
        line_list = line.strip().split(delimiter)
        line_list = [x.strip() for x in line_list if x != ""]  # Remove multiple delimiter
        if len(line_list) == 1 and num == 0 and comment == 0:
            # Start new conformer and set line counts to read.
            num = int(line_list[0])
            comment = 1
        elif comment > 0:
            # Comment comes before atom block and must always be read.
            comment_list.append(str(line))
            comment = 0
        elif num > 0:
            if len(line_list) <= 1:
                logging.error("Expected to read atom-coordinate block but got comment or line count instead.")
            atoms.append(str(line_list[0]).lower().capitalize())
            coordinates.append([float(x) for x in line_list[1:]])
            if num == 1:
                # This was last line for this conformer. Append result and reset current list.
                mol_list.append([atoms, coordinates])
                num = 0
                atoms = []
                coordinates = []
            else:
                # Finished reading a atom line.
                num = num - 1
        else:
            logging.warning("Empty line in xyz file for mismatch in atom count found.")
    # close file
    infile.close()
    return mol_list


def write_mol_block_list_to_sdf(mol_block_list, filepath):
    """Write a list of mol blocks as string into a SDF file.

    Args:
        mol_block_list (list): List of mol blocks as string.
        filepath (str): File path for SDF file.

    Returns:
        None.
    """
    with open(filepath, "w+") as file:
        for i, mol_block in enumerate(mol_block_list):
            if mol_block is not None:
                file.write(mol_block)
                if i < len(mol_block_list) - 1:
                    file.write("$$$$\n")
            else:
                file.write("".join(["\n",
                                    "     FAIL\n",
                                    "\n",
                                    "  0  0  0  0  0  0  0  0  0  0 V2000\n",
                                    "M  END\n"]))
                if i < len(mol_block_list) - 1:
                    file.write("$$$$\n")


def read_mol_list_from_sdf_file(filepath, line_by_line=False):
    """Simple loader to load a SDF file by only splitting.

    Args:
        filepath (str): File path for SDF file.
        line_by_line (bool): Whether to read SDF file line by line.

    Returns:
        list: List of mol blocks as string.
    """
    mol_list = []
    with open(filepath, "r") as f:
        if not line_by_line:
            all_sting = f.read()
            mol_list = all_sting.split("$$$$\n")
        else:
            iter_mol = ""
            for line in f:
                if line == "$$$$\n":
                    mol_list.append(iter_mol)
                    iter_mol = ""
                else:
                    iter_mol = iter_mol + line
            if iter_mol != "":
                mol_list.append(iter_mol)
    # Check if there was tailing $$$$ with nothing to follow.
    # Split will make empty string at the end, which does not match actual number of mol blocks.
    if len(mol_list[-1]) == 0:
        mol_list = mol_list[:-1]
    return mol_list


def read_smiles_file(file_path):
    """Simply python function to read smiles from file.

    Args:
        file_path (str): File path for smiles file.

    Returns:
        list: List of smiles.
    """
    with open(file_path, "r") as f:
        smile_list = f.readlines()
    return smile_list


def write_smiles_file(file_path, smile_list):
    """Simply python function to write smiles to file.

    Args:
        file_path (str): File path for smiles file.
        smile_list (list): List of smiles to write to file.

    Returns:
        None
    """
    with open(file_path, "w+") as f:
        for i, x in enumerate(smile_list):
            if i == len(smile_list)-1:
                f.write(x)
            else:
                f.write(x + "\n")



import os
import pickle
import numpy as np
import json

from kgcnn.data.qm import QMDataset
from kgcnn.data.download import DownloadDataset
from kgcnn.mol.io import write_list_to_xyz_file


class QM9Dataset(QMDataset, DownloadDataset):
    """Store and process QM9 dataset."""

    download_info = {
        "dataset_name": "QM9",
        "data_directory_name": "qm9",
        # https://ndownloader.figshare.com/files/3195398
        "download_url": "https://ndownloader.figshare.com/files/3195389",
        "download_file_name": 'dsgdb9nsd.xyz.tar.bz2',
        "unpack_tar": True,
        "unpack_zip": False,
        "unpack_directory_name": 'dsgdb9nsd.xyz',
    }

    def __init__(self, reload: bool = False, verbose: int = 1):
        """Initialize QM9 dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 0=silent. Default is 1.
        """
        QMDataset.__init__(self, verbose=verbose, dataset_name="QM9")
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        self.label_names = ['A', 'B', 'C', 'mu', 'alpha', 'homo', 'lumo', 'gap', 'r2', 'zpve', 'U0', 'U', 'H',
                             'G', 'Cv']
        self.label_units = ["GHz", "GHz", "GHz", "D", r"a_0^3", "eV", "eV", "eV", r"a_0^2", "eV", "eV", "eV", "eV",
                             "eV", r"cal/mol K"]
        self.label_unit_conversion = np.array(
            [[1.0, 1.0, 1.0, 1.0, 1.0, 27.2114, 27.2114, 27.2114, 1.0, 27.2114, 27.2114, 27.2114,
              27.2114, 27.2114, 1.0]]
        )  # Pick always same units for training
        self.dataset_name = "QM9"
        self.require_prepare_data = True
        self.fits_in_memory = True
        self.verbose = verbose
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name)
        self.file_name = "qm9.xyz"

        if self.require_prepare_data:
            self.prepare_data(overwrite=reload)

        if self.fits_in_memory:
            self.read_in_memory()

    def prepare_data(self, overwrite: bool = False, xyz_column_name: str = None, make_sdf: bool = True):
        """Process data by loading all single xyz-files and store all pickled information to file.
        The single files are deleted afterwards, requires to re-extract the tar-file for overwrite.

        Args:
            overwrite (bool): Whether to redo the processing, requires un-zip of the data again. Defaults to False.
            xyz_column_name (str): Not used.
            make_sdf (bool): Whether to make SDF file.
        """
        path = self.data_directory

        dataset_size = 133885

        if (os.path.exists(os.path.join(path, "qm9.pickle")) or os.path.exists(
                os.path.join(path, "qm9.json"))) and not overwrite:
            self.info("Single molecules already pickled.")
        else:
            if not os.path.exists(os.path.join(path, 'dsgdb9nsd.xyz')):
                self.error("Can not find extracted dsgdb9nsd.xyz directory. Run reload dataset again.")
                return
            qm9 = []
            # Read individual files
            self.info("Reading dsgdb9nsd files ...")
            for i in range(1, dataset_size + 1):
                mol = []
                file = "dsgdb9nsd_" + "{:06d}".format(i) + ".xyz"
                open_file = open(os.path.join(path, "dsgdb9nsd.xyz", file), "r")
                lines = open_file.readlines()
                mol.append(int(lines[0]))
                labels = lines[1].strip().split(' ')[1].split('\t')
                if int(labels[0]) != i:
                    self.warning("Index for QM9 not matching xyz-file.")
                labels = [lines[1].strip().split(' ')[0].strip()] + [int(labels[0])] + [float(x) for x in labels[1:]]
                mol.append(labels)
                cords = []
                for j in range(int(lines[0])):
                    atom_info = lines[2 + j].strip().split('\t')
                    cords.append([atom_info[0]] + [float(x.replace('*^', 'e')) for x in atom_info[1:]])
                mol.append(cords)
                freqs = lines[int(lines[0]) + 2].strip().split('\t')
                freqs = [float(x) for x in freqs]
                mol.append(freqs)
                smiles = lines[int(lines[0]) + 3].strip().split('\t')
                mol.append(smiles)
                inchis = lines[int(lines[0]) + 4].strip().split('\t')
                mol.append(inchis)
                open_file.close()
                qm9.append(mol)

            # Save pickle data
            self.info("Saving qm9.json ...")
            with open(os.path.join(path, "qm9.json"), 'w') as f:
                json.dump(qm9, f)

            # Remove file after reading
            self.info("Cleaning up extracted files...")
            for i in range(1, dataset_size + 1):
                file = "dsgdb9nsd_" + "{:06d}".format(i) + ".xyz"
                file = os.path.join(path, "dsgdb9nsd.xyz", file)
                os.remove(file)

        if os.path.exists(os.path.join(path, self.file_name)) and not overwrite:
            self.info("Single xyz-file %s for molecules already created." % self.file_name)
        else:
            self.info("Reading dataset...")
            if os.path.exists(os.path.join(path, "qm9.pickle")):
                with open(os.path.join(path, "qm9.pickle"), 'rb') as f:
                    qm9 = pickle.load(f)
            elif os.path.exists(os.path.join(path, "qm9.json")):
                with open(os.path.join(path, "qm9.json"), 'rb') as f:
                    qm9 = json.load(f)
            else:
                raise FileNotFoundError("Can not find pickled QM9 dataset.")

            # Try extract bond-info and save mol-file.
            self.info("Writing single xyz-file ...")
            pos = [[y[1:] for y in x[2]] for x in qm9]
            atoms = [[y[0] for y in x[2]] for x in qm9]
            atoms_pos = [[x, y] for x, y in zip(atoms, pos)]
            write_list_to_xyz_file(os.path.join(path, "qm9.xyz"), atoms_pos)

        super(QM9Dataset, self).prepare_data(overwrite=overwrite)
        return self

    def read_in_memory(self, label_column_name: str = None):
        """Load the pickled QM9 data into memory and already split into items.

        Args:
            label_column_name(str): Not used.

        Returns:
            self
        """
        path = self.data_directory

        self.info("Reading dataset ...")
        if os.path.exists(os.path.join(path, "qm9.pickle")):
            with open(os.path.join(path, "qm9.pickle"), 'rb') as f:
                qm9 = pickle.load(f)
        elif os.path.exists(os.path.join(path, "qm9.json")):
            with open(os.path.join(path, "qm9.json"), 'rb') as f:
                qm9 = json.load(f)
        else:
            raise FileNotFoundError("Can not find pickled QM9 dataset.")

        # labels
        # self.length = 133885
        labels = np.array([x[1][1:] if len(x[1]) == 17 else x[1] for x in qm9])  # Remove 'gdb' tag here

        # Atoms as nodes
        atoms = [[y[0] for y in x[2]] for x in qm9]
        atom_dict = {'H': 1, 'C': 6, 'N': 7, 'O': 8, 'F': 9}
        zval = [[atom_dict[y] for y in x] for x in atoms]
        outzval = [np.array(x, dtype="int") for x in zval]
        nodes = outzval

        # Mean molecular weight mmw
        massdict = {'H': 1.0079, 'C': 12.0107, 'N': 14.0067, 'O': 15.9994, 'F': 18.9984}
        mass = [[massdict[y] for y in x] for x in atoms]
        mmw = np.array([[np.mean(x), len(x)] for x in mass])

        # Coordinates
        coord = [[[y[1], y[2], y[3]] for y in x[2]] for x in qm9]
        coord = [np.array(x) for x in coord]

        # Labels
        ids = [x for x in labels[:, 0]]
        actual_labels = self.label_unit_conversion*labels[:, 1:]
        actual_labels = [x for x in actual_labels]

        self.assign_property("graph_number", ids)
        self.assign_property("node_coordinates", coord)
        self.assign_property("graph_labels", actual_labels)
        self.assign_property("node_symbol", atoms)
        self.assign_property("node_number", nodes)
        self.assign_property("graph_attributes", [x for x in mmw])

        # Try to read mol information
        self.read_in_memory_sdf()
        return self


# from kgcnn.data.qm import QMGraphLabelScaler
# dataset = QM9Dataset()
# scaler = QMGraphLabelScaler([{"class_name": "StandardScaler", "config": {}},
#                              {"class_name": "ExtensiveMolecularScaler", "config": {}},
#                              {"class_name": "ExtensiveMolecularScaler", "config": {}}])
# trafo_labels = scaler.fit_transform(np.array(dataset.graph_labels)[:, :3], dataset.node_number)
# rev_labels = scaler.inverse_transform(dataset.node_number, tafo_labels
# print(np.amax(np.abs(dataset.graph_labels-rev_labels)))



import tensorflow as tf
import logging
from typing import Union


logging.basicConfig()  # Module logger
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


def set_devices_gpu(gpu_ids_list: Union[list, int], restrict_memory: bool = True):
    r"""Set the visible devices from a list of GPUs. Used to assign a process to a separate GPU.

    .. note::
        Important is to restrict memory growth since a single tensorflow process will allocate almost all
        GPU memory, so two fits can not run on same GPU.

    Args:
        gpu_ids_list (list): Device list.
        restrict_memory (bool): Whether to restrict memory growth. Default is True.

    Returns:
        None.
    """
    if gpu_ids_list is None:
        return

    if isinstance(gpu_ids_list, int):
        gpu_ids_list = [gpu_ids_list]

    if len(gpu_ids_list) <= 0:
        module_logger.info("No gpu to set")
        return

    if tf.test.is_built_with_gpu_support() is False and tf.test.is_built_with_cuda() is False:
        module_logger.warning("No cuda support")
        module_logger.warning("Can not set GPU")
        return

    try:
        gpus = tf.config.list_physical_devices('GPU')
    except:
        module_logger.error("Can not get device list, do nothing")
        return

    if isinstance(gpus, list):
        if len(gpus) <= 0:
            module_logger.warning("No devices found")
            module_logger.warning("Can not set GPU")
            return
        try:
            gpus_used = [gpus[i] for i in gpu_ids_list if 0 <= i < len(gpus)]
            tf.config.set_visible_devices(gpus_used, 'GPU')
            module_logger.info("Setting visible devices: %s" % gpus_used)
            if restrict_memory:
                for gpu in gpus_used:
                    module_logger.info("Restrict Memory: %s" % gpu.name)
                    tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.experimental.list_logical_devices("GPU")
            module_logger.info("Physical GPUS: %s, Logical GPUS: %s" % (len(gpus), len(logical_gpus)))
        except RuntimeError as e:
            # Visible devices must be set before GPUs have been initialized
            print(e)



import tensorflow as tf
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.ops.partition import partition_row_indexing
from kgcnn.ops.axis import get_positive_axis
ks = tf.keras


@ks.utils.register_keras_serializable(package='kgcnn', name='GatherEmbedding')
class GatherEmbedding(GraphBaseLayer):
    r"""Gather node or edge embedding from an index list.

    The embeddings are gather from a ragged index tensor. An edge is defined by index tuple :math:`(i ,j)`.
    In the default definition, index :math:`i` is expected to be the receiving or target node.
    Effectively, the layer simply does:

    .. code-block:: python

        tf.gather(embedding, indices, batch_dims=1, axis=1)

    Additionally, the gathered embeddings can be split or concatenated along the index dimension after gather,
    by setting :obj:`split_axis` or :obj:`concat_axis` if index shape is known during build.

    .. note:
        Default of this layer is concatenation with :obj:`concat_axis=2`.

    Example of usage for :obj:`GatherEmbedding`:

    .. code-block:: python

        nodes = tf.ragged.constant([[[0.0],[1.0]],[[2.0],[3.0],[4.0]]], ragged_rank=1)
        edge_idx = tf.ragged.constant([[[0,1],[1,0]],[[0,2],[1,2]]], ragged_rank=1)
        print(GatherEmbedding()([nodes, edge_idx]))
    """

    def __init__(self,
                 axis: int = 1,
                 concat_axis: int = 2,
                 split_axis: int = None,
                 **kwargs):
        r"""Initialize layer.

        Args:
            axis (int): The axis to gather embeddings from. Default is 1.
            concat_axis (int): The axis which concatenates embeddings. Default is 2.
            split_axis (int): The axis to split the gathered embeddings. Default is None.
        """
        super(GatherEmbedding, self).__init__(**kwargs)
        self.concat_axis = concat_axis
        self.axis = axis
        self.split_axis = split_axis

        if split_axis is not None and concat_axis is not None:
            raise ValueError("Can not both split and concatenate new index axis. At least one must be `None`.")

    def build(self, input_shape):
        super(GatherEmbedding, self).build(input_shape)

        if len(input_shape) != 2:
            print("Number of inputs for layer %s is expected to be 2." % self.name)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [embeddings, tensor_index]

                - embeddings (tf.RaggedTensor): Node embeddings of shape `(batch, [N], F)`
                - tensor_index (tf.RaggedTensor): Edge indices referring to nodes of shape `(batch, [M], 2)`

        Returns:
            tf.RaggedTensor: Gathered node embeddings that match the number of edges of shape `(batch, [M], 2*F)`
        """
        # The primary case for aggregation of nodes from node feature list. Case from doc-string.
        # Faster implementation via values and indices shifted by row-partition. Equal to disjoint implementation.
        if all([isinstance(x, tf.RaggedTensor) for x in inputs]):
            if all([x.ragged_rank == 1 for x in inputs]) and self.axis == 1 and self.concat_axis in [None, 2] and self.split_axis in [None, 2]:
                node, node_part = inputs[0].values, inputs[0].row_splits
                edge_index, edge_part = inputs[1].values, inputs[1].row_lengths()
                disjoint_list = partition_row_indexing(edge_index, node_part, edge_part,
                                                       partition_type_target="row_splits",
                                                       partition_type_index="row_length", to_indexing='batch',
                                                       from_indexing=self.node_indexing)
                out = tf.gather(node, disjoint_list, axis=0)
                if self.concat_axis == 2:
                    out = tf.concat([out[:, i] for i in range(edge_index.shape[1])], axis=1)
                    return tf.RaggedTensor.from_row_lengths(out, edge_part, validate=self.ragged_validate)
                if self.split_axis == 2:
                    return [tf.RaggedTensor.from_row_lengths(out[:, i], edge_part, validate=self.ragged_validate) for i
                            in range(edge_index.shape[1])]
                return tf.RaggedTensor.from_row_lengths(out, edge_part, validate=self.ragged_validate)

        # For arbitrary gather from ragged tensor use tf.gather with batch_dims=1.
        # Works in tf.__version__>=2.4 now!
        out = tf.gather(inputs[0], inputs[1], batch_dims=1, axis=self.axis)
        if self.concat_axis is not None:
            out = tf.concat([tf.gather(out, i, axis=self.concat_axis) for i in range(out.shape[self.concat_axis])],
                            axis=self.concat_axis)
        if self.split_axis is not None:
            out = [tf.gather(out, i, axis=self.split_axis) for i in range(out.shape[self.split_axis])]
        return out

    def get_config(self):
        """Update layer config."""
        config = super(GatherEmbedding, self).get_config()
        config.update({"concat_axis": self.concat_axis, "axis": self.axis, "split_axis": self.split_axis})
        return config


GatherNodes = GatherEmbedding


@ks.utils.register_keras_serializable(package='kgcnn', name='GatherEmbeddingSelection')
class GatherEmbeddingSelection(GraphBaseLayer):
    r"""Gather node or edge embedding for a defined index in the index list.

    The embeddings are gather from a ragged index tensor for a list of specific indices which are given by
    :obj:`selection_index`. This can be used for ingoing or outgoing nodes or angles.
    Returns a list of embeddings for each :obj:`selection_index`. An edge is defined by index tuple :math:`(i, j)`.
    In the default definition, index :math:`i` is expected to be the receiving or target node.
    Effectively, the layer simply does:

    .. code-block:: python

        tf.gather(embedding, indices[:, :, selection_index], batch_dims=1, axis=1)

    Additionally, the axis for gather can be specified for target and index tensor via :obj:`axis` and
    :obj:`axis_indices`.
    This layer always returns a list of embeddings even if :obj:`selection_index` is of type :obj:`int`.

    Example of usage for :obj`GatherEmbeddingSelection`:

    .. code-block:: python

        nodes = tf.ragged.constant([[[0.0],[1.0]],[[2.0],[3.0],[4.0]]], ragged_rank=1)
        edge_idx = tf.ragged.constant([[[0,1],[1,0]],[[0,2],[1,2]]], ragged_rank=1)
        print(GatherEmbeddingSelection([0, 1])([nodes, edge_idx]))

    """

    def __init__(self, selection_index, axis: int = 1, axis_indices: int = 2, **kwargs):
        r"""Initialize layer.

        Args:
            selection_index (list, int): Which indices to gather embeddings for.
            axis (int): Axis to gather embeddings from. Default is 1.
            axis_indices (int): From which axis to take the indices for gather. Default is 2.
        """
        super(GatherEmbeddingSelection, self).__init__(**kwargs)
        self.axis = axis
        self.axis_indices = axis_indices

        if not isinstance(selection_index, (list, tuple, int)):
            raise ValueError("Indices for selection must be list or tuple for layer `GatherEmbeddingSelection`.")

        if isinstance(selection_index, int):
            self.selection_index = [selection_index]
        else:
            self.selection_index = list(selection_index)

    def build(self, input_shape):
        """Build layer."""
        super(GatherEmbeddingSelection, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [embeddings, tensor_index]

                - embeddings (tf.RaggedTensor): Node embeddings of shape `(batch, [N], F)`
                - tensor_index (tf.RaggedTensor): Edge indices referring to nodes of shape `(batch, [M], 2)`

        Returns:
            list: Gathered node embeddings matching the number of edges of shape `(batch, [M], F)` for selection_index.
        """
        # The primary case for aggregation of nodes from node feature list. Case from doc-string.
        # Faster implementation via values and indices shifted by row-partition. Equal to disjoint implementation.
        if all([isinstance(x, tf.RaggedTensor) for x in inputs]):
            if all([x.ragged_rank == 1 for x in inputs]) and self.axis == 1 and self.axis_indices == 2:
                # We cast to values here
                node, node_part = inputs[0].values, inputs[0].row_splits
                edge_index, edge_part = inputs[1].values, inputs[1].row_lengths()
                indexlist = partition_row_indexing(edge_index, node_part, edge_part,
                                                   partition_type_target="row_splits",
                                                   partition_type_index="row_length",
                                                   to_indexing='batch',
                                                   from_indexing=self.node_indexing)
                out = [tf.gather(node, tf.gather(indexlist, i, axis=1), axis=0) for i in self.selection_index]
                out = [tf.RaggedTensor.from_row_lengths(x, edge_part, validate=self.ragged_validate) for x in out]
                return out

        # For arbitrary gather from ragged tensor use tf.gather with batch_dims=1.
        out = [tf.gather(inputs[0], tf.gather(inputs[1], i, axis=self.axis_indices), batch_dims=1, axis=self.axis) for i
               in self.selection_index]  # Works in tf.__version__>=2.4
        return out

    def get_config(self):
        """Update layer config."""
        config = super(GatherEmbeddingSelection, self).get_config()
        config.update({"axis": self.axis, "axis_indices": self.axis_indices, "selection_index": self.selection_index})
        return config


GatherNodesSelection = GatherEmbeddingSelection


@ks.utils.register_keras_serializable(package='kgcnn', name='GatherNodesIngoing')
class GatherNodesIngoing(GatherEmbeddingSelection):
    r"""Gather receiving or ingoing nodes of edges with index :math:`i`.

    An edge is defined by index tuple :math:`(i, j)`.
    In the default definition, index :math:`i` is expected to be the receiving or target node.
    The layer inherits from :obj:`GatherEmbeddingSelection` and effectively does:

    .. code-block:: python

        GatherEmbeddingSelection(selection_index=0)(inputs)[0]

    """

    def __init__(self, selection_index: int = 0, **kwargs):
        r"""Initialize layer.

        Args:
            selection_index (list, int): Which index to gather embeddings for. Default is 0.
        """
        super(GatherNodesIngoing, self).__init__(selection_index=selection_index, **kwargs)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [nodes, tensor_index]

                - nodes (tf.RaggedTensor): Node embeddings of shape `(batch, [N], F)`
                - tensor_index (tf.RaggedTensor): Edge indices referring to nodes of shape `(batch, [M], 2)`

        Returns:
            tf.RaggedTensor: Gathered node embeddings for ingoing nodes of edges of shape `(batch, [M], F)`
        """
        return super(GatherNodesIngoing, self).call(inputs, **kwargs)[0]


@ks.utils.register_keras_serializable(package='kgcnn', name='GatherNodesOutgoing')
class GatherNodesOutgoing(GatherEmbeddingSelection):
    r"""Gather sending or outgoing nodes of edges with index :math:`j`.

    An edge is defined by index tuple :math:`(i, j)`.
    In the default definition, index :math:`j` is expected to be the sending or source node.
    The layer inherits from :obj:`GatherEmbeddingSelection` and effectively does:

    .. code-block:: python

        GatherEmbeddingSelection(selection_index=1)(inputs)[0]

    """

    def __init__(self, selection_index: int = 1, **kwargs):
        r"""Initialize layer.

        Args:
            selection_index (list, int): Which index to gather embeddings for. Default is 1.
        """
        super(GatherNodesOutgoing, self).__init__(selection_index=selection_index, **kwargs)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [nodes, tensor_index]

                - nodes (tf.RaggedTensor): Node embeddings of shape `(batch, [N], F)`
                - tensor_index (tf.RaggedTensor): Edge indices referring to nodes of shape `(batch, [M], 2)`

        Returns:
            tf.RaggedTensor: Gathered node embeddings for outgoing nodes of edges of shape `(batch, [M], F)`
        """
        return super(GatherNodesOutgoing, self).call(inputs, **kwargs)[0]


@ks.utils.register_keras_serializable(package='kgcnn', name='GatherState')
class GatherState(GraphBaseLayer):
    r"""Layer to repeat environment or global state for a specific embeddings (ragged) tensor like node or edge lists.
    
    To repeat the correct global state (like an environment feature vector) for each sample,
    a tensor with the target shape or length, partition in case of ragged target tensors is required.

    Mostly used to concatenate a global state :math:`\mathbf{s}` with node embeddings :math:`\mathbf{h}_i`
    like for example:

    .. math::
        \mathbf{h}_i = \mathbf{h}_i \oplus \mathbf{s}

    where this layer only repeats :math:`\mathbf{s}` to match a ragged embedding tensor :math:`\mathbf{h}_i`.

    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(GatherState, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        super(GatherState, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [state, target]

                - state (tf.Tensor): Graph specific embedding tensor. This is tensor of shape `(batch, F)`
                - target (tf.RaggedTensor): Target to collect state for [N] of shape `(batch, [N], F)`

        Returns:
            tf.RaggedTensor: Graph embedding with repeated single state for each graph of shape `(batch, [N], F)`.
        """
        env = inputs[0]
        dyn_inputs = inputs[1]

        if isinstance(dyn_inputs, tf.RaggedTensor):
            target_len = dyn_inputs.row_lengths()
        else:
            target_len = tf.repeat(tf.shape(dyn_inputs)[1], tf.shape(dyn_inputs)[0])

        out = tf.repeat(env, target_len, axis=0)
        out = tf.RaggedTensor.from_row_lengths(out, target_len, validate=self.ragged_validate)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(GatherState, self).get_config()
        return config



import numpy as np
import scipy as sp
import scipy.special
import tensorflow as tf
from scipy.optimize import brentq


@tf.function
def tf_spherical_bessel_jn_explicit(x, n=0):
    r"""Compute spherical bessel functions :math:`j_n(x)` for constant positive integer :math:`n` explicitly.
    TensorFlow has to cache the function for each :math:`n`. No gradient through :math:`n` or very large number
    of :math:`n`'s is possible.
    The spherical bessel functions and there properties can be looked up at
    https://en.wikipedia.org/wiki/Bessel_function#Spherical_Bessel_functions.
    For this implementation the explicit expression from https://dlmf.nist.gov/10.49 has been used.
    The definition is:

    :math:`a_{k}(n+\tfrac{1}{2})=\begin{cases}\dfrac{(n+k)!}{2^{k}k!(n-k)!},&k=0,1,\dotsc,n\\
    0,&k=n+1,n+2,\dotsc\end{cases}`

    :math:`\mathsf{j}_{n}\left(z\right)=\sin\left(z-\tfrac{1}{2}n\pi\right)\sum_{k=0}^{\left\lfloor n/2\right\rfloor}
    (-1)^{k}\frac{a_{2k}(n+\tfrac{1}{2})}{z^{2k+1}}+\cos\left(z-\tfrac{1}{2}n\pi\right)
    \sum_{k=0}^{\left\lfloor(n-1)/2\right\rfloor}(-1)^{k}\frac{a_{2k+1}(n+\tfrac{1}{2})}{z^{2k+2}}.`

    Args:
        x (tf.Tensor): Values to compute :math:`j_n(x)` for.
        n (int): Positive integer for the bessel order :math:`n`.

    Returns:
        tf.Tensor: Spherical bessel function of order :math:`n`
    """
    sin_x = tf.sin(x - n * np.pi / 2)
    cos_x = tf.cos(x - n * np.pi / 2)
    sum_sin = tf.zeros_like(x)
    sum_cos = tf.zeros_like(x)
    for k in range(int(np.floor(n / 2)) + 1):
        if 2 * k < n + 1:
            prefactor = float(sp.special.factorial(n + 2 * k) / np.power(2, 2 * k) / sp.special.factorial(
                2 * k) / sp.special.factorial(n - 2 * k) * np.power(-1, k))
            sum_sin += prefactor * tf.pow(x, - (2 * k + 1))
    for k in range(int(np.floor((n - 1) / 2)) + 1):
        if 2 * k + 1 < n + 1:
            prefactor = float(sp.special.factorial(n + 2 * k + 1) / np.power(2, 2 * k + 1) / sp.special.factorial(
                2 * k + 1) / sp.special.factorial(n - 2 * k - 1) * np.power(-1, k))
            sum_cos += prefactor * tf.pow(x, - (2 * k + 2))
    return sum_sin * sin_x + sum_cos * cos_x


@tf.function
def tf_spherical_bessel_jn(x, n=0):
    r"""Compute spherical bessel functions :math:`j_n(x)` for constant positive integer :math:`n` via recursion.
    TensorFlow has to cache the function for each :math:`n`. No gradient through :math:`n` or very large number
    of :math:`n` is possible.
    The spherical bessel functions and there properties can be looked up at
    https://en.wikipedia.org/wiki/Bessel_function#Spherical_Bessel_functions.
    The recursive rule is constructed from https://dlmf.nist.gov/10.51. The recursive definition is:

    :math:`j_{n+1}(z)=((2n+1)/z)j_{n}(z)-j_{n-1}(z)`

    :math:`j_{0}(x)=\frac{\sin x}{x}`

    :math:`j_{1}(x)=\frac{1}{x}\frac{\sin x}{x} - \frac{\cos x}{x}`

    :math:`j_{2}(x)=\left(\frac{3}{x^{2}} - 1\right)\frac{\sin x}{x} - \frac{3}{x}\frac{\cos x}{x}`

    Args:
        x (tf.Tensor): Values to compute :math:`j_n(x)` for.
        n (int): Positive integer for the bessel order :math:`n`.

    Returns:
        tf.tensor: Spherical bessel function of order :math:`n`
    """
    if n < 0:
        raise ValueError("Order parameter must be >= 0 for this implementation of spherical bessel function.")
    if n == 0:
        return tf.sin(x) / x
    elif n == 1:
        return tf.sin(x) / tf.square(x) - tf.cos(x) / x
    else:
        j_n = tf.sin(x) / x
        j_nn = tf.sin(x) / tf.square(x) - tf.cos(x) / x
        for i in range(1, n):
            temp = j_nn
            j_nn = (2 * i + 1) / x * j_nn - j_n
            j_n = temp
        return j_nn


@tf.function
def tf_legendre_polynomial_pn(x, n=0):
    r"""Compute the (non-associated) Legendre polynomial :math:`P_n(x)` for constant positive integer :math:`n`
    via explicit formula.
    TensorFlow has to cache the function for each :math:`n`. No gradient through :math:`n` or very large number
    of :math:`n` is possible.
    Closed form can be viewed at https://en.wikipedia.org/wiki/Legendre_polynomials.

    :math:`P_n(x)=\sum_{k=0}^{\lfloor n/2\rfloor} (-1)^k \frac{(2n - 2k)! \, }{(n-k)! \, (n-2k)! \, k! \, 2^n} x^{n-2k}`

    Args:
        x (tf.Tensor): Values to compute :math:`P_n(x)` for.
        n (int): Positive integer for :math:`n` in :math:`P_n(x)`.

    Returns:
        tf.tensor: Legendre Polynomial of order :math:`n`.
    """
    out_sum = tf.zeros_like(x)
    prefactors = [
        float((-1) ** k * sp.special.factorial(2 * n - 2 * k) / sp.special.factorial(n - k) / sp.special.factorial(
            n - 2 * k) / sp.special.factorial(k) / 2 ** n) for k in range(0, int(np.floor(n / 2)) + 1)]
    powers = [float(n - 2 * k) for k in range(0, int(np.floor(n / 2)) + 1)]
    for i in range(len(powers)):
        out_sum = out_sum + prefactors[i] * tf.pow(x, powers[i])
    return out_sum


@tf.function
def tf_spherical_harmonics_yl(theta, l=0):
    r"""Compute the spherical harmonics :math:`Y_{ml}(\cos\theta)` for :math:`m=0` and constant non-integer :math:`l`.
    TensorFlow has to cache the function for each :math:`l`. No gradient through :math:`l` or very large number
    of :math:`n` is possible. Uses a simplified formula with :math:`m=0` from
    https://en.wikipedia.org/wiki/Spherical_harmonics:

    :math:`Y_{l}^{m}(\theta ,\phi)=\sqrt{\frac{(2l+1)}{4\pi} \frac{(l -m)!}{(l +m)!}} \, P_{l}^{m}(\cos{\theta }) \,
    e^{i m \phi}`

    where the associated Legendre polynomial simplifies to :math:`P_l(x)` for :math:`m=0`:

    :math:`P_n(x)=\sum_{k=0}^{\lfloor n/2\rfloor} (-1)^k \frac{(2n - 2k)! \, }{(n-k)! \, (n-2k)! \, k! \, 2^n} x^{n-2k}`

    Args:
        theta (tf.Tensor): Values to compute :math:`Y_l(\cos\theta)` for.
        l (int): Positive integer for :math:`l` in :math:`Y_l(\cos\theta)`.

    Returns:
        tf.tensor: Spherical harmonics for :math:`m=0` and constant non-integer :math:`l`.
    """
    x = tf.cos(theta)
    out_sum = tf.zeros_like(x)
    prefactors = [
        float((-1) ** k * sp.special.factorial(2 * l - 2 * k) / sp.special.factorial(l - k) / sp.special.factorial(
            l - 2 * k) / sp.special.factorial(k) / 2 ** l) for k in range(0, int(np.floor(l / 2)) + 1)]
    powers = [float(l - 2 * k) for k in range(0, int(np.floor(l / 2)) + 1)]
    for i in range(len(powers)):
        out_sum = out_sum + prefactors[i] * tf.pow(x, powers[i])
    out_sum = out_sum * float(np.sqrt((2 * l + 1) / 4 / np.pi))
    return out_sum


@tf.function
def tf_associated_legendre_polynomial(x, l=0, m=0):
    r"""Compute the associated Legendre polynomial :math:`P_{l}^{m}(x)` for :math:`m` and constant positive
    integer :math:`l` via explicit formula.
    Closed Form from taken from https://en.wikipedia.org/wiki/Associated_Legendre_polynomials.

    :math:`P_{l}^{m}(x)=(-1)^{m}\cdot 2^{l}\cdot (1-x^{2})^{m/2}\cdot \sum_{k=m}^{l}\frac{k!}{(k-m)!}\cdot x^{k-m}
    \cdot \binom{l}{k}\binom{\frac{l+k-1}{2}}{l}`.

    Args:
        x (tf.Tensor): Values to compute :math:`P_{l}^{m}(x)` for.
        l (int): Positive integer for :math:`l` in :math:`P_{l}^{m}(x)`.
        m (int): Positive/Negative integer for :math:`m` in :math:`P_{l}^{m}(x)`.

    Returns:
        tf.tensor: Legendre Polynomial of order n.
    """
    if np.abs(m) > l:
        raise ValueError("Error: Legendre polynomial must have -l<= m <= l")
    if l < 0:
        raise ValueError("Error: Legendre polynomial must have l>=0")
    if m < 0:
        m = -m
        neg_m = float(np.power(-1, m) * sp.special.factorial(l - m) / sp.special.factorial(l + m))
    else:
        neg_m = 1

    x_prefactor = tf.pow(1 - tf.square(x), m / 2) * float(np.power(-1, m) * np.power(2, l))
    sum_out = tf.zeros_like(x)
    for k in range(m, l + 1):
        sum_out += tf.pow(x, k - m) * float(
            sp.special.factorial(k) / sp.special.factorial(k - m) * sp.special.binom(l, k) *
            sp.special.binom((l + k - 1) / 2, l))

    return sum_out * x_prefactor * neg_m


def spherical_bessel_jn(r, n):
    r"""Compute spherical Bessel function :math:`j_n(r)` via scipy.
    The spherical bessel functions and there properties can be looked up at
    https://en.wikipedia.org/wiki/Bessel_function#Spherical_Bessel_functions .

    Args:
        r (np.ndarray): Argument
        n (np.ndarray): Order.

    Returns:
        np.array: Values of the spherical Bessel function
    """
    return np.sqrt(np.pi / (2 * r)) * sp.special.jv(n + 0.5, r)


def spherical_bessel_jn_zeros(n, k):
    r"""Compute the first :math:`k` zeros of the spherical bessel functions :math:`j_n(r)` up to
    order :math:`n` (excluded).
    Taken from the original implementation of DimeNet at https://github.com/klicperajo/dimenet.

    Args:
        n: Order.
        k: Number of zero crossings.

    Returns:
        np.ndarray: List of zero crossings of shape (n, k)
    """
    zerosj = np.zeros((n, k), dtype="float32")
    zerosj[0] = np.arange(1, k + 1) * np.pi
    points = np.arange(1, k + n) * np.pi
    racines = np.zeros(k + n - 1, dtype="float32")
    for i in range(1, n):
        for j in range(k + n - 1 - i):
            foo = brentq(spherical_bessel_jn, points[j], points[j + 1], (i,))
            racines[j] = foo
        points = racines
        zerosj[i][:k] = racines[:k]

    return zerosj


def spherical_bessel_jn_normalization_prefactor(n, k):
    r"""Compute the normalization or rescaling pre-factor for the spherical bessel functions :math:`j_n(r)` up to
    order :math:`n` (excluded) and maximum frequency :math:`k` (excluded).
    Taken from the original implementation of DimeNet at https://github.com/klicperajo/dimenet.

    Args:
        n: Order.
        k: frequency.

    Returns:
        np.ndarray: Normalization of shape (n, k)
    """
    zeros = spherical_bessel_jn_zeros(n, k)
    normalizer = []
    for order in range(n):
        normalizer_tmp = []
        for i in range(k):
            normalizer_tmp += [0.5 * spherical_bessel_jn(zeros[order, i], order + 1) ** 2]
        normalizer_tmp = 1 / np.array(normalizer_tmp) ** 0.5
        normalizer += [normalizer_tmp]
    return np.array(normalizer)



import unittest
import tempfile
import os
import random
from pprint import pprint

import numpy as np
import pandas as pd
from rdkit import RDLogger
RDLogger.DisableLog('rdApp.*')

from kgcnn.data.moleculenet import MoleculeNetDataset
from kgcnn.mol.encoder import OneHotEncoder


SIMPLE_SMILES_CSV = """
index,name,label,smiles
1,Propanolol,1,[Cl].CC(C)NCC(O)COc1cccc2ccccc12
2,Terbutylchlorambucil,1,C(=O)(OC(C)(C)C)CCCc1ccc(cc1)N(CCCl)CCCl
3,40730,1,c12c3c(N4CCN(C)CC4)c(F)cc1c(c(C(O)=O)cn2C(C)CO3)=O
4,24,1,C1CCN(CC1)Cc1cccc(c1)OCCCNC(=O)C
"""


FAULTY_SMILES_CSV = """
index,name,label,smiles
57,compound 36,1,CN(C)Cc1ccc(CSCCNC2=C([N+]([O-])=O)C(Cc3ccccc3)=CN2)o1
58,19,0,CNC(=NC#N)Nc1cccc(c1)c1csc(n1)N=C(N)N
59,Y-G 14,1,n(ccc1)c(c1)CCNC
60,15,1,O=N([O-])C1=C(CN=C1NCCSCc2ncccc2)Cc3ccccc3
"""


class TestMoleculeNetDataset(unittest.TestCase):

    # MoleculeNetDataset needs a folder in which it can operate. We create a fixture which will set up a new temp
    # folder for each test case.

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.temp_dir = None
        self.temp_path = None
        self.file_name = 'test.csv'

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.temp_path = self.temp_dir.__enter__()

    def tearDown(self):
        self.temp_dir.__exit__(None, None, None)

    # This is a utility method. It will create a CSV file as the source for the dataset in the current temporary folder
    # using the string CSV content
    def write_string(self, string: str):
        with open(os.path.join(self.temp_path, self.file_name), mode='w') as file:
            file.write(string)

    # ~ ACTUAL TEST CASES

    def test_construction_basically_works(self):
        self.write_string(SIMPLE_SMILES_CSV)
        molnet = MoleculeNetDataset(data_directory=self.temp_path, file_name=self.file_name, dataset_name='test')
        self.assertIsInstance(molnet, MoleculeNetDataset)

    def test_basically_works(self):
        self.write_string(SIMPLE_SMILES_CSV)
        molnet = MoleculeNetDataset(data_directory=self.temp_path, file_name=self.file_name, dataset_name='test')

        molnet.prepare_data(
            overwrite=False,
            smiles_column_name='smiles'
        )

        molnet.read_in_memory(
            label_column_name='label',
            add_hydrogen=False,
        )
        # We know that this csv data has 4 entries
        self.assertEqual(len(molnet), 4)

        # Basic tests for one of the entries
        data = molnet[random.randint(0, 3)]
        self.assertIsInstance(data, dict)
        # These are the basic fields which every entry is supposed to have and all of them need to be
        # numpy arrays
        keys = ['graph_size', 'graph_labels', 'node_coordinates', 'node_number',
                'node_symbol', 'edge_indices']
        for key in keys:
            self.assertIn(key, data)
            print(type(data[key]))
            self.assertIsInstance(data[key], np.ndarray)

    def test_setting_attributes_works(self):
        self.write_string(SIMPLE_SMILES_CSV)
        molnet = MoleculeNetDataset(data_directory=self.temp_path, file_name=self.file_name, dataset_name='test')

        # This is list the keys which the dataset as a whole should contain after they have been read
        # to the memory
        default_keys = ['graph_size', 'graph_labels', 'node_coordinates', 'node_number',
                        'node_symbol', 'edge_indices']
        # This is the list of keys, which the data set should contain after the "set_attributes" has been
        # called
        attribute_keys = ['graph_attributes', 'node_attributes', 'edge_attributes']

        molnet.prepare_data(
            overwrite=False,
            smiles_column_name='smiles'
        )

        molnet.read_in_memory(
            label_column_name='label',
            add_hydrogen=False,
        )
        pprint(molnet.__dict__.keys())

        data = molnet[random.randint(0, 3)]
        # At this point, the default keys should be present but not the attribute keys
        for key in default_keys:
            self.assertIn(key, data)

        for key in attribute_keys:
            self.assertNotIn(key, data)

        molnet.set_attributes(
            nodes=['Symbol'],
            encoder_nodes={'Symbol': OneHotEncoder(['C', 'O'], dtype='str', add_unknown=True)},
            edges=['BondType'],
            encoder_edges={'BondType': int}
        )

        # We know that this csv data has 4 entries
        self.assertEqual(len(molnet), 4)

        data = molnet[random.randint(0, 3)]
        # Now also the attribute keys should be present.
        for key in default_keys:
            self.assertIn(key, data)

        for key in attribute_keys:
            self.assertIn(key, data)

    def test_faulty_smiles_do_not_cause_exception(self):
        self.write_string(FAULTY_SMILES_CSV)
        molnet = MoleculeNetDataset(data_directory=self.temp_path, file_name=self.file_name, dataset_name='test')

        molnet.prepare_data(
            overwrite=False,
            smiles_column_name='smiles'
        )

        molnet.read_in_memory(
            label_column_name='label',
            add_hydrogen=False,
        )

        # At this point the test is successful if there was no error...
        self.assertEqual(len(molnet), 4)

    def test_pandas_from_csv(self):
        """
        Simply test if pandas data frame indexing works as expected.
        """
        self.write_string(SIMPLE_SMILES_CSV)
        data = pd.read_csv(os.path.join(self.temp_path, self.file_name))

        row_index = 0
        # This line is supposed to return a dict for the values of the row of the CSV file with the specified
        # index
        row_dict = dict(data.loc[row_index])
        self.assertEqual(len(row_dict), 4)
        self.assertIn('index', row_dict.keys())
        self.assertIn('label', row_dict.keys())
        self.assertIn('smiles', row_dict.keys())
        self.assertIn('name', row_dict.keys())

        self.assertFalse(True)

    def test_map_molecule_attributes_basically_works(self):
        """
        Tests if the new method "._map_molecule_callbacks" works on a simple example
        """
        self.write_string(SIMPLE_SMILES_CSV)
        molnet = MoleculeNetDataset(data_directory=self.temp_path, file_name=self.file_name, dataset_name='test')

        callbacks = {
            'name': lambda mg, dd: np.array(dd['name'], dtype='str'),
            'label': lambda mg, dd: np.array(dd['label'], dtype='str')
        }

        # Before we can call the method, the data needs to be prepared because we need to make sure that the molecule
        # file exists before attempting to read from it.
        molnet.prepare_data(
            overwrite=False,
            smiles_column_name='smiles'
        )
        # This method will automatically add properties with the string key names of the "callbacks" dict to the
        # underlying GraphList based on the transformations defined by the callback functions.
        molnet._map_molecule_callbacks(callbacks)

        molecule = molnet[1]
        self.assertIn('name', molecule)
        self.assertIn('label', molecule)

        # But the other default attributes are not part of it, since we did not specify those in the
        # callbacks dict
        self.assertNotIn('node_symbol', molecule)

    def test_graph_labels_is_one_dimensional_array(self):
        self.write_string(SIMPLE_SMILES_CSV)

        # In this first setting we pass a single string as the label name. The corresponding "graph_labels"
        # attribute of the dataset elements should be a one-dim np array regardless (and not zero dim)
        molnet = MoleculeNetDataset(data_directory=self.temp_path, file_name=self.file_name, dataset_name='test')
        molnet.prepare_data()
        molnet.read_in_memory(label_column_name='label')
        molnet.set_attributes()

        molecule = molnet[0]
        self.assertIsInstance(molecule['graph_labels'], np.ndarray)
        self.assertEqual(len(molecule['graph_labels'].shape), 1)

        # But it should also be possible to pass a list of column names in to get a true np array of
        # multiple labels
        molnet = MoleculeNetDataset(data_directory=self.temp_path, file_name=self.file_name, dataset_name='test')
        molnet.prepare_data()
        molnet.read_in_memory(label_column_name=['label', 'name'])
        molnet.set_attributes()

        molecule = molnet[0]
        self.assertIsInstance(molecule['graph_labels'], np.ndarray)
        self.assertEqual(len(molecule['graph_labels'].shape), 1)
        self.assertEqual(len(molecule['graph_labels']), 2)
        # Testing if it is the correct value as well
        self.assertEqual(molecule['graph_labels'][1], 'Propanolol')



import warnings
from copy import deepcopy
import numpy as np
from scipy.spatial import Voronoi
from networkx import MultiDiGraph
from pyxtal import pyxtal
from pymatgen.core.structure import Structure
from pymatgen.core.periodic_table import Element
from pymatgen.optimization.neighbors import find_points_in_spheres
from typing import Union

def get_symmetrized_graph(structure: Union[Structure, pyxtal]) -> MultiDiGraph:

    graph = MultiDiGraph()
    if isinstance(structure, pyxtal):
        pyxtal_cell = structure 
    elif isinstance(structure, Structure):
        try:
            pyxtal_cell = pyxtal()
            pyxtal_cell.from_seed(structure)
        except:
            # use trivial spacegroup (with spacegroup number == 1)
            # if spglib isn't able to calculate symmetries
            frac_coords = np.array([site.frac_coords for site in structure.sites])
            frac_coords = _to_unit_cell(frac_coords)
            for node_idx, site in enumerate(structure.sites):
                graph.add_node(node_idx, atomic_number=site.specie.number,
                        asymmetric_mapping=node_idx,
                        frac_coords=frac_coords[node_idx],
                        coords=site.coords,
                        symmop=np.eye(4),
                        multiplicity=1)
            setattr(graph, 'lattice_matrix', structure.lattice.matrix)
            setattr(graph, 'spacegroup', 1)
            return graph
    else:
        ValueError("This method takes either a pymatgen.core.structure.Structure or a pyxtal object.")

    atomic_numbers, frac_coords, asymmetric_mapping, symmops, multiplicities = [], [], [], [], []
    for site in pyxtal_cell.atom_sites:
        atomic_numbers += (site.multiplicity * [Element(site.specie).Z])
        asymmetric_mapping += (site.multiplicity * [len(asymmetric_mapping)])
        frac_coords.append(site.coords)
        symmops += [symmop.affine_matrix for symmop in site.wp.ops]
        multiplicities += (site.multiplicity * [site.multiplicity])
    frac_coords = _to_unit_cell(np.vstack(frac_coords))
    lattice = pyxtal_cell.lattice.matrix
    coords = frac_coords @ lattice
    for node_idx in range(len(atomic_numbers)):
        graph.add_node(node_idx, atomic_number=atomic_numbers[node_idx],
                asymmetric_mapping=asymmetric_mapping[node_idx],
                frac_coords=frac_coords[node_idx],
                coords=coords[node_idx],
                symmop= symmops[node_idx],
                multiplicity=multiplicities[node_idx])
    setattr(graph, 'lattice_matrix', lattice)
    setattr(graph, 'spacegroup', pyxtal_cell.group.number)
    
    return graph

def structure_to_empty_graph(structure: Union[Structure, pyxtal], symmetrize=False) -> MultiDiGraph:

    if symmetrize:
        return get_symmetrized_graph(structure)
    else:
        if isinstance(structure, pyxtal):
            structure = structure.to_pymatgen()
        graph = MultiDiGraph()
        frac_coords = np.array([site.frac_coords for site in structure.sites])
        frac_coords = _to_unit_cell(frac_coords)
        for node_idx, site in enumerate(structure.sites):
            graph.add_node(node_idx, atomic_number=site.specie.number,
                    frac_coords=frac_coords[node_idx],
                    coords=site.coords)
        setattr(graph, 'lattice_matrix', structure.lattice.matrix)
        return graph

def add_knn_bonds(graph: MultiDiGraph, k=12, max_radius=10., inplace=False):

    lattice = deepcopy(graph.lattice_matrix)
    frac_coords = np.array([data[1] for data in graph.nodes(data='frac_coords')])
    coords = frac_coords @ lattice
    index1, index2, offset_vectors, distances = find_points_in_spheres(coords,
                                                                   coords,
                                                                   r=max_radius,
                                                                   pbc=np.array([True] * 3, dtype=int),
                                                                   lattice=lattice,
                                                                   tol=1e-8)
    # Remove self_loops:
    no_self_loops = np.argwhere(~np.isclose(distances, 0)).reshape(-1)
    index1 = index1[no_self_loops]
    index2 = index2[no_self_loops]
    offset_vectors = offset_vectors[no_self_loops]
    distances = distances[no_self_loops]
    
    new_graph = graph if inplace else deepcopy(graph)

    for node_idx in range(new_graph.number_of_nodes()):
        idxs = np.argwhere(index1 == node_idx)[:,0]
        sorted_idxs = idxs[np.argsort(distances[idxs])]
        sorted_idxs = sorted_idxs[:k]
        # If the max_radius doesn't capture 12 neighbors, try again with double the max_radius
        if len(sorted_idxs) < k:
            return add_knn_bonds(new_graph, k=k, max_radius=max_radius*2, inplace=True)
        for edge_idx in sorted_idxs:
            new_graph.add_edge(index2[edge_idx], index1[edge_idx], cell_translation=offset_vectors[edge_idx],\
                    distance=distances[edge_idx])

    return new_graph



def add_radius_bonds(graph: MultiDiGraph, radius=5., inplace=False):
    new_graph = graph if inplace else deepcopy(graph)

    lattice = deepcopy(graph.lattice_matrix)
    frac_coords = np.array([data[1] for data in graph.nodes(data='frac_coords')])
    coords = frac_coords @ lattice
    index1, index2, offset_vectors, distances = find_points_in_spheres(coords,
                                                                   coords,
                                                                   r=radius,
                                                                   pbc=np.array([True] * 3, dtype=int),
                                                                   lattice=lattice,
                                                                   tol=1e-8)
    # Remove self_loops:
    no_self_loops = np.argwhere(~np.isclose(distances, 0)).reshape(-1)
    index1 = index1[no_self_loops]
    index2 = index2[no_self_loops]
    offset_vectors = offset_vectors[no_self_loops]
    distances = distances[no_self_loops]

    if len(index1) == 0:
        warnings.warn('No edges added to the graph, \
consider increasing the radius and check your graph input instance.')

    new_graph = deepcopy(graph)
    for source, target, cell_translation, dist in zip(index2, index1, offset_vectors, distances):
        new_graph.add_edge(source, target, cell_translation=cell_translation, distance=dist)

    return new_graph


def add_voronoi_bonds(graph: MultiDiGraph, inplace=False):
    new_graph = graph if inplace else deepcopy(graph)

    lattice = graph.lattice_matrix
    frac_coords = np.array([data[1] for data in graph.nodes(data='frac_coords')])
    dim = lattice.shape[0]
    assert dim == 3
    size = np.array([1,1,1])
    expanded_frac_coords = _get_super_cell_frac_coords(lattice, frac_coords, size)
    expanded_coords = expanded_frac_coords @ lattice
    flattened_expanded_coords = expanded_coords.reshape(-1, dim)

    voronoi = Voronoi(flattened_expanded_coords)
    ridge_points_unraveled = np.array(np.unravel_index(voronoi.ridge_points, expanded_coords.shape[:-1]))
    # shape: (num_ridges, 2 (source, target), 4 (3 cell_index + 1 atom_index)) 
    ridge_points_unraveled = np.moveaxis(ridge_points_unraveled, [0,1,2], [2,0,1])

    # Filter ridges that have source in the centered unit cell
    source_in_center_cell = np.argwhere(np.all(ridge_points_unraveled[:,0,:dim] == 1, axis=-1))[:,0]
    # Filter ridges that have target in the centered unit cell
    target_in_center_cell = np.argwhere(np.all(ridge_points_unraveled[:,1,:dim] == 1, axis=-1))[:,0]
    
    edge_info = np.vstack(
            [ridge_points_unraveled[source_in_center_cell][:,[1,0]], ridge_points_unraveled[target_in_center_cell]])
    
    cell_translations = (edge_info[:,0,:-1] - size).astype(float)
    edge_indices = edge_info[:,:,-1]

    distances = []
    for i in range(len(edge_indices)):
        d = np.linalg.norm(expanded_coords[tuple(edge_info[i][0])] - expanded_coords[tuple(edge_info[i][1])])
        distances.append(d)
    
    for nodes, cell_translation, dist in zip(edge_indices, cell_translations, distances):
        source, target = nodes[0], nodes[1]
        new_graph.add_edge(source, target, cell_translation=cell_translation, distance=dist)
    
    return new_graph

def remove_duplicate_edges(graph: MultiDiGraph, inplace=False) -> MultiDiGraph:
    new_graph = graph if inplace else deepcopy(graph)
    edge_counter = set()
    remove_edges = set()
    for e in new_graph.edges(data='cell_translation', keys=True):
        id_ = (e[0], e[1]) + tuple(e[3].astype(int))
        if id_ in edge_counter:
            remove_edges.add((e[0], e[1], e[2]))
        else:
            edge_counter.add(id_)
    for edge in remove_edges:
        new_graph.remove_edge(edge[0], edge[1], key=edge[2])
    return new_graph


def get_mesh(size, dim):
    if isinstance(size, int):
        size = [size] * dim
    else:
        size = list(size)
    assert len(size) == dim
    
    mesh = np.array(np.meshgrid(*tuple([np.arange(i) for i in size])))
    mesh = np.moveaxis(mesh, [0,1], [-1,1])
    return mesh

def get_cube(dim: int):
    return get_mesh(2, dim)

def _get_max_diameter(lattice):
    dim = lattice.shape[0]
    cube = get_cube(dim)
    max_radius = np.max(np.linalg.norm((cube - 1/2) @ lattice, axis=1))
    return max_radius * 2

def _get_super_cell_size_from_radius(lattice: np.ndarray, radius: float):
    dim = lattice.shape[0]
    max_diameter = _get_max_diameter(lattice)
    radius_ = radius + max_diameter
    cell_indices = np.ceil(np.sum(np.abs(np.linalg.inv(lattice)), axis=0) * radius_).astype(int)
    cells = get_mesh(cell_indices + 1, dim)
    lattice_point_coords = cells @ lattice
    images = np.argwhere(np.linalg.norm(lattice_point_coords, axis=-1) <= radius_)
    super_cell_size = images.max(axis=0)
    return super_cell_size

def _get_super_cell_frac_coords(lattice, frac_coords, size):
    dim = lattice.shape[0]
    
    if isinstance(size, int):
        size = [size] * dim
    else:
        size = list(size)
    assert len(size) == dim
    
    doubled_size = np.array(size) * 2 + 1
    mesh = get_mesh(doubled_size, dim)
    # frac_coords_expanded.shape == (1,1,1,num_atoms,3) (for dim == 3)
    frac_coords_expanded = np.expand_dims(frac_coords, np.arange(dim).tolist())
    # mesh_expanded.shape == (double_size[0], double_size[1], double_size[2], 1, 3) (for dim == 3)
    mesh_expanded = np.expand_dims(mesh - size, -2)
    expanded_frac_coords = mesh_expanded + frac_coords_expanded
    
    return expanded_frac_coords



def reshape_at_axis(arr, axis, new_shape):
    # move reshape axis to last axis position, because np.reshape reshapes this first
    arr_tmp = np.moveaxis(arr, axis, -1)
    shape = arr_tmp.shape[:-1] + new_shape
    new_positions = np.arange(len(new_shape)) + axis
    old_positions = np.arange(len(new_shape)) + (len(arr.shape) - 1)
    # now call np.reshape and move axis to right position
    return np.moveaxis(arr_tmp.reshape(shape), old_positions, new_positions)

def pairwise_diff(coords1, coords2):
    # Difference calculated at last axis of both inputs
    assert coords1.shape[-1] == coords2.shape[-1]
    coords1_reshaped = coords1.reshape(-1,coords1.shape[-1])
    coords2_reshaped = coords2.reshape(-1,coords2.shape[-1])
    diffs = np.expand_dims(coords2_reshaped, 0) - np.expand_dims(coords1_reshaped, 1)
    return reshape_at_axis(reshape_at_axis(diffs, 1, coords2.shape[:-1]), 0, coords1.shape[:-1])

def add_edge_information(graph: MultiDiGraph, inplace=False,
        frac_offset=False, offset=True, distance=True) -> MultiDiGraph:

    new_graph = graph if inplace else deepcopy(graph)
    if graph.number_of_edges() == 0:
        return new_graph

    add_frac_offset = frac_offset
    add_offset = offset
    add_distance = distance

    frac_coords1 = []
    frac_coords2 = []
    cell_translations = []
    
    # Collect necessary coordinate informations for calculations
    for e in graph.edges(data='cell_translation'):
        frac_coords1.append(graph.nodes[e[0]]['frac_coords'])
        cell_translations.append(e[2])
        frac_coords2.append(graph.nodes[e[1]]['frac_coords'])

    # Do calculations in vectorized form (instead of doing it inside the edge loop)
    frac_coords1 = np.array(frac_coords1)
    frac_coords2 = np.array(frac_coords2)
    cell_translations = np.array(cell_translations)
    frac_offset = frac_coords2 - (frac_coords1 + cell_translations)
    offset = frac_offset @ graph.lattice_matrix
    if add_distance:
        distances = np.linalg.norm(offset, axis=-1)

    # Add calculated informations to edge attributes
    for i, e in enumerate(graph.edges(data=True)):
        if add_frac_offset:
            e[2]['frac_offset'] = frac_offset[i]
        if add_offset:
            e[2]['offset'] = offset[i]
        if add_distance:
            e[2]['distance'] = distances[i]

    return new_graph

def to_supercell_graph(graph: MultiDiGraph, size):

    supercell_graph = MultiDiGraph()
    size_ = list(size) + [graph.number_of_nodes()]
    new_num_nodes = np.prod(size_)
    for node in range(new_num_nodes):
        idx = np.unravel_index(node, size_)
        cell_translation = idx[:3]
        node_num = idx[3]
        data = deepcopy(graph.nodes[node_num])
        data['frac_coords'] = data['frac_coords'] + np.array(cell_translation)
        data['coords'] = data['frac_coords'] @ graph.lattice_matrix
        supercell_graph.add_node(node, **data)
        
    for edge in graph.edges(data=True):
        for cell_idx in range(np.prod(size)):
            cell_translation1 = np.unravel_index(cell_idx, size)
            cell_translation2 = (edge[2]['cell_translation'] + np.array(cell_translation1)).astype(int)
            if np.all(cell_translation2 >= 0) and np.all(cell_translation2 < size):
                new_source = np.ravel_multi_index(list(cell_translation2) + [edge[0]], size_)
                new_target = np.ravel_multi_index(list(cell_translation1) + [edge[1]], size_)
                data = deepcopy(edge[2])
                supercell_graph.add_edge(new_source, new_target, **data)
        
    setattr(supercell_graph, 'lattice_matrix', graph.lattice_matrix)
    if hasattr(graph, 'spacegroup'):
        setattr(supercell_graph, 'spacegroup', graph.spacegroup)
    return supercell_graph


def to_asymmetric_unit_graph(graph: MultiDiGraph) -> MultiDiGraph:

    asymmetric_mapping = np.array([node[1] for node in graph.nodes(data='asymmetric_mapping')])
    if None in asymmetric_mapping:
        raise ValueError("Graph does not contain symmetry informations. \
Make sure to create the graph with `structure_to_empty_graph` with the `symmetrize` \
argument set to `True`.")
    asu_node_indice, inv_asymmetric_mapping = np.unique(asymmetric_mapping, return_inverse=True)
    
    asu_graph = MultiDiGraph()
    setattr(asu_graph, 'lattice_matrix', graph.lattice_matrix)
    setattr(asu_graph, 'spacegroup', graph.spacegroup)
    new_nodes_idx = {}

    # Add nodes of asymmetric unit to asu_graph
    for i, node_idx in enumerate(asu_node_indice):
        new_nodes_idx[node_idx] = i
        data = deepcopy(graph.nodes[node_idx])
        data['unit_cell_index'] = data['asymmetric_mapping']
        del data['asymmetric_mapping']
        del data['symmop']
        asu_graph.add_node(i, **data)

    if graph.number_of_edges() == 0:
        return asu_graph

    edges_to_keep = graph.in_edges(asu_node_indice, data=True)
    for e in edges_to_keep:
        source_data = graph.nodes[e[0]]
        new_source_idx = inv_asymmetric_mapping[e[0]]
        new_target_idx = new_nodes_idx[e[1]]
        data = deepcopy(e[2])
        data['symmop'] = source_data['symmop']
        asu_graph.add_edge(new_source_idx, new_target_idx, **data)

    return asu_graph

def _to_unit_cell(frac_coords):
    return frac_coords % 1. % 1.

